@@ -13,3 +13,7 @@ void _wn_set_fgbg PROTO((int is_fg, char *sred, char *sgreen, char *sblue));
 void _wn_init_sunview_colors PROTO((swin_t *w, int is_mainwin));
 
 extern int wn__Use_mono;
+extern int wn__Use_alloc_color;
+extern int wn__Use_alloc_color_for_highlight;
+extern int wn__Use_menu_colors;
+extern int wn__Use_extended_text_colors;
