@@ -7,20 +7,26 @@
  *  unaltered copy of this message (including the copyright).
  */
 
-char wn_wn_event_c_sccsid[] = "@(#)wn_event.c	1.39 1/7/92 (UKC)";
+char wn_wn_event_c_sccsid[] = "@(#)wn_event.c	1.46 25 May 1995 (UKC)";
 
-/*  Some versions of the select() macros use bzero, so map it to memset
- */
-#define bzero(s, count)		memset(s, 0, count)
-
-#ifdef SVR4
-#include <sys/fcntl.h>
+#if defined(__svr4__) || defined(__SVR4)
+#define SVR4 1
 #endif
 
-#include <sys/types.h>
 #include <sys/time.h>
-#include <sys/file.h>
+#include <sys/types.h>
+#include <fcntl.h>
 #include <signal.h>
+
+#if defined(SVR4)
+#define USE_STREAMS
+#endif
+
+#ifdef USE_STREAMS
+#include <stropts.h>
+#include <sys/conf.h>
+#endif
+
 #include <stdio.h>
 #include <errno.h>
 #include <string.h>
@@ -30,10 +36,9 @@ char wn_wn_event_c_sccsid[] = "@(#)wn_event.c	1.39 1/7/92 (UKC)";
 #include <unistd.h>
 #endif
 
-#include <local/ukcprog.h>
-
 #include "wn.h"
 #include "wn_priv.h"
+#include "wn_init.h"  /* RCB, for process_client_message */
 #include "wn_sc.h"
 #include "wn_win.h"
 #include "wn_misc.h"
@@ -44,6 +49,12 @@ char wn_wn_event_c_sccsid[] = "@(#)wn_event.c	1.39 1/7/92 (UKC)";
 #include <X11/keysym.h>
 #endif
 
+#if defined(__sun__) && defined(SVR4) && !defined(SA_RESTART)
+#define SA_RESTART	0x00000004
+#endif
+
+typedef void (*Sigfunc)PROTO((int));
+
 static void get_event PROTO((int wn, event_t *ev, int *p_xyset));
 static int select_on_mask PROTO((long win_mask, int sample, long *p_resmask));
 static void set_event_win PROTO((event_t *ev));
@@ -59,17 +70,41 @@ static void sun_to_wn_event PROTO((Event *sun_event, event_t *ev));
 
 #ifdef X11
 static bool event_wanted PROTO((Display *display, XEvent *x_event, char *arg));
+static void set_event_fields PROTO((event_t *ev, int x, int y,
+				    unsigned xbstate, unsigned xbflipped));
 #endif
+static Sigfunc safe_signal PROTO((int sig, Sigfunc func));
 
 /*  Space for a single pushed back event.
  *  See wn_pushback_event() and wn_next_event().
  */
 static event_t Pushedback_event;
 
+/*  Timeout when doing input in sample mode.  Set by wn_set_input_timeout().
+ *  Default is no timeout (e.g. wait indefinitely).
+ */
+static struct timeval *Request_mode_timeout = NULL;
+
 #ifdef X11
 /*  See extern declaration in wn_event.h.
  */
 Window wn__Last_event_window;
+
+/* FIXME: need to go through the stupid contortions to find out which
+ *        modifier ALT is on this keyboard.
+ */
+static unsigned Alt_mask = Mod1Mask;
+
+/*  Multi click stuff by Russ Browne (RCB),
+**  Applied MicroSystems Corp.
+**  russ@amc.com
+*/
+
+#define DEFAULT_MULTICLICK_TIME 250	/* Match xterm default */
+static int multiClickTime = DEFAULT_MULTICLICK_TIME;
+static int numClicks;
+static long lastButtonTime;
+static int  lastButtonsFlags;
 #endif
 
 /*  BUG: this typedef means the user can only set the first 32 fds.
@@ -119,6 +154,29 @@ fdmask_t mask;
 	User_fd_mask = mask;
 }
 
+long
+wn_set_input_timeout(msec)
+long msec;
+{
+	static long last_msec = 0;
+	long res;
+
+	if (msec == 0) {
+		Request_mode_timeout = NULL;
+	}
+	else {
+		static struct timeval tvbuf;
+
+		tvbuf.tv_sec = msec / 1000;
+		tvbuf.tv_usec = (msec % 1000) * 1000;
+		Request_mode_timeout = &tvbuf;
+	}
+
+	res = last_msec;
+	last_msec = msec;
+	return res;
+}
+
 static int
 select_on_mask(win_mask, sample, p_resmask)
 fdmask_t win_mask;
@@ -127,7 +185,7 @@ fdmask_t *p_resmask;
 {
 	static int first_call = TRUE;
 	static struct timeval zero_timeout = { 0, 0 };
-	register struct timeval *timeout;
+	struct timeval *timeout;
 	fd_set mask;
 	int sel_res;
 
@@ -137,9 +195,11 @@ fdmask_t *p_resmask;
 		first_call = FALSE;
 	}
 		
-	timeout = sample ? &zero_timeout : NULL;
+	timeout = sample ? &zero_timeout : Request_mode_timeout;
 
 	for (;;) {
+		struct timeval tvbuf, *stm;
+
 		FD_ZERO(&mask);
 		mask.fds_bits[0] = win_mask | User_fd_mask;
 #ifdef SUNVIEW
@@ -147,13 +207,25 @@ fdmask_t *p_resmask;
 			FD_SET(Wakeup_fd, &mask);
 #endif
 
-		sel_res = select(32, &mask, (fd_set *)NULL, (fd_set *)NULL, timeout);
+		/*  Beware of select modifying the timeout parameter.
+		 */
+	 	if (timeout != NULL) {
+			 tvbuf = *timeout;
+			 stm = &tvbuf;
+		}
+		else {
+			stm = timeout;
+		}
+			
+		sel_res = select(32, &mask,(fd_set *)NULL,(fd_set *)NULL, stm);
 		switch(sel_res) {
 		case -1:
 			if (errno != EINTR)
 				wn__panic("select failed in wn_next_event");
 			return EV_INTERRUPT;
 		case 0:
+			if (timeout == Request_mode_timeout)
+				return EV_INTERRUPT;
 			return EV_MOUSE_MOVED;
 		default:
 #ifdef SUNVIEW
@@ -270,6 +342,7 @@ register event_t *ev;
 			ev->ev_char = sun_event->ie_code;
 		break;
 	}
+	
 	if (flipped != 0) {
 		ev->ev_flags = flipped;
 		if (win_inputnegevent(sun_event)) {
@@ -281,11 +354,11 @@ register event_t *ev;
 			ev->ev_buttons |= flipped;
 		}
 
-		ev->ev_buttons &= ~(B_SHIFT_KEY | B_CONTROL_KEY);
+		ev->ev_buttons &= ~(B_SHIFT_MASK | B_CONTROL_MASK);
 		if (event_shift_is_down(sun_event))
-			ev->ev_buttons |= B_SHIFT_KEY;
+			ev->ev_buttons |= B_SHIFT_MASK;
 		if (event_ctrl_is_down(sun_event))
-			ev->ev_buttons |= B_CONTROL_KEY;
+			ev->ev_buttons |= B_CONTROL_MASK;
 	}
 }
 
@@ -391,6 +464,29 @@ _wn_enable_self_deiconise()
 #ifdef X11
 #define SAMPLE_MASK	((REQUEST_MASK) & ~(PointerMotionMask))
 
+void
+wn_set_MultiClickTime()
+{
+   /*
+   **   We recognize both
+   **	  MultiClickTime     = time in ms, as in xterm
+   **	  MultiClickTimeout  = time in 1/10 sec, as in openwin
+   */
+   int scale = 1;
+   int timeMs = 0;
+   const char* timeSpec = wn_get_default("MultiClickTime");
+   if ( timeSpec == NULL )
+   {
+	timeSpec = wn_get_default("MultiClickTimeout");
+	scale = 100;
+   }
+   if ( timeSpec )
+	timeMs = scale*atoi(timeSpec);
+
+   if ( timeMs > 0 )
+     multiClickTime = timeMs;		/* RCB */
+}
+
 static bool
 event_wanted(display, x_event, arg)
 Display *display;
@@ -412,8 +508,9 @@ int *p_xyset;
 {
 	static swin_t *last_w;
 	XEvent x_event;
+	XKeyEvent *xk;
+	XButtonEvent *xb;
 	int sample;
-	int xbstate, xbflipped, buttons, flipped, modflags;
 	int old_width, old_height, xbuttons, nbytes;
 	char ch;
 	char *keystr;
@@ -564,86 +661,78 @@ int *p_xyset;
 		ev->ev_type = EV_MOUSE_MOVED;
 		break;
 	case ButtonPress:
+		/* Monitor multiple clicks on ButtonPress
+		** Then fall into ButtonRelease case ( RCB)
+		*/
+		xb = (XButtonPressedEvent *)&x_event;
+		if (   numClicks < 3
+		    && (xb->time - lastButtonTime)
+			< multiClickTime)
+			numClicks++;
+		else
+			numClicks = 0;
 	case ButtonRelease:
-		ev->ev_type = (x_event.type == ButtonPress) ?
-						EV_BUTTON_DOWN : EV_BUTTON_UP;
-		ev->ev_x = ((XButtonPressedEvent *)&x_event)->x;
-		ev->ev_y = ((XButtonPressedEvent *)&x_event)->y;
-		*p_xyset = TRUE;
-		xbstate = ((XButtonPressedEvent *)&x_event)->state;
-		xbflipped = ((XButtonPressedEvent *)&x_event)->button;
+		xb = (XButtonPressedEvent *)&x_event;
+		lastButtonTime = xb->time;
 
-		buttons = 0;
-		if (xbstate & Button1Mask)
-			buttons |= B_LEFT;
-		if (xbstate & Button2Mask)
-			buttons |= B_MIDDLE;
-		if (xbstate & Button3Mask)
-			buttons |= B_RIGHT;
-
-		modflags = 0;
-		if (xbstate & ShiftMask)
-			modflags |= B_SHIFT_KEY;
-		if (xbstate & ControlMask)
-			modflags |= B_CONTROL_KEY;
-
-		switch(xbflipped) {
-		case Button1:
-			flipped = B_LEFT;
-			break;
-		case Button2:
-			flipped = B_MIDDLE;
-			break;
-		case Button3:
-			flipped = B_RIGHT;
-			break;
-		default:
-			flipped = 0;
+		if (xb->button == 0){
 			ev->ev_type = EV_MOUSE_MOVED;
-			break;
 		}
-		ev->ev_buttons = (buttons ^ flipped) | modflags;
-		ev->ev_flags = flipped;
+		else {
+			ev->ev_type = (x_event.type == ButtonPress) ?
+						EV_BUTTON_DOWN : EV_BUTTON_UP;
+		}
+		set_event_fields(ev, xb->x, xb->y, xb->state, xb->button);
+		*p_xyset = TRUE;
+		/*
+		**  If the buttons changed, clear the click counter.
+		**  The OR it into the ev_buttons field.
+		*/
+		if ( ev->ev_flags != lastButtonsFlags)
+		   numClicks = 0;
+
+		lastButtonsFlags = ev->ev_flags;
+		ev->ev_buttons |= (numClicks<<8);
 		break;
 	case KeyPress:
+		xk = (XKeyEvent *)&x_event;
 		keystr = &ch;
-		nbytes = XLookupString((XKeyPressedEvent *)&x_event, keystr, 1,
+		nbytes = XLookupString(xk, keystr, 1,
 					(KeySym *)NULL, (XComposeStatus *)NULL);
 		if (nbytes == 0) {
-			int keysym;
+			KeySym keysym;
 
-			keysym = XLookupKeysym((XKeyEvent *)&x_event, 0);
+			keysym = XLookupKeysym(xk, 0);
 			nbytes = 1;
 			switch (keysym) {
-				case XK_Left:	ch = WN_CH_LEFT_ARROW;	break;
-				case XK_Right:	ch = WN_CH_RIGHT_ARROW;	break;
-				case XK_Up:	ch = WN_CH_UP_ARROW;	break;
-				case XK_Down:	ch = WN_CH_DOWN_ARROW;	break;
-				case XK_F1:	ch = WN_CH_FUNCKEY(1);	break;
-				case XK_F2:	ch = WN_CH_FUNCKEY(2);	break;
-				case XK_F3:	ch = WN_CH_FUNCKEY(3);	break;
-				case XK_F4:	ch = WN_CH_FUNCKEY(4);	break;
-				case XK_F5:	ch = WN_CH_FUNCKEY(5);	break;
-				case XK_F6:	ch = WN_CH_FUNCKEY(6);	break;
-				case XK_F7:	ch = WN_CH_FUNCKEY(7);	break;
-				case XK_F8:	ch = WN_CH_FUNCKEY(8);	break;
-				case XK_F9:	ch = WN_CH_FUNCKEY(9);	break;
-				case XK_F10:	ch = WN_CH_FUNCKEY(10);	break;
-				case XK_F11:	ch = WN_CH_FUNCKEY(11);	break;
-				case XK_F12:	ch = WN_CH_FUNCKEY(12);	break;
-				default:	nbytes = 0;		break;
+			case XK_Left:
+				ch = WN_CH_LEFT_ARROW;
+				break;
+			case XK_Right:
+				ch = WN_CH_RIGHT_ARROW;
+				break;
+			case XK_Up:
+				ch = WN_CH_UP_ARROW;
+				break;
+			case XK_Down:
+				ch = WN_CH_DOWN_ARROW;
+				break;
+			default:
+				if (keysym >= XK_F1 && keysym <= XK_F35)
+					ch = WN_CH_FUNCKEY((keysym + 1)-XK_F1);
+				else
+					nbytes = 0;
+				break;
 			}
 		}
 				
-		ev->ev_type = (nbytes > 0) ? EV_KEY : EV_OTHER;
+		ev->ev_type = (nbytes > 0) ? EV_KEY : 0;
+		
 		if (nbytes > 0) /* BUG: what if nbytes > 1 ? */
 			ev->ev_char = *keystr;
 
-		if (((XKeyEvent *)&x_event)->state & ShiftMask)
-			ev->ev_buttons |= B_SHIFT_KEY;
-		if (((XKeyEvent *)&x_event)->state & ControlMask)
-			ev->ev_buttons |= B_CONTROL_KEY;
-
+		set_event_fields(ev, xk->x, xk->y, xk->state, 0);
+		*p_xyset = TRUE;
 		break;
 
 	case SelectionRequest:
@@ -655,11 +744,60 @@ int *p_xyset;
 		ev->ev_type = EV_LOST_SELECTION;
 		break;
 
+	case ClientMessage:
+		process_client_message(&x_event, w->w_win);
+		ev->ev_type = EV_MOUSE_MOVED;
+		break;
+    
 	default:
 		ev->ev_type = 0;
 	}
 }
 
+static void
+set_event_fields(ev, x, y, xbstate, xbflipped)
+event_t *ev;
+int x, y;
+unsigned xbstate, xbflipped;
+{
+	int buttons, flipped, modflags;
+
+	buttons = 0;
+	if (xbstate & Button1Mask)
+		buttons |= B_LEFT;
+	if (xbstate & Button2Mask)
+		buttons |= B_MIDDLE;
+	if (xbstate & Button3Mask)
+		buttons |= B_RIGHT;
+
+	modflags = 0;
+	if (xbstate & ShiftMask)
+		modflags |= B_SHIFT_MASK;
+	if (xbstate & ControlMask)
+		modflags |= B_CONTROL_MASK;
+	if (xbstate & Alt_mask)
+		modflags |= B_ALT_MASK;
+
+	switch(xbflipped) {
+	case Button1:
+		flipped = B_LEFT;
+		break;
+	case Button2:
+		flipped = B_MIDDLE;
+		break;
+	case Button3:
+		flipped = B_RIGHT;
+		break;
+	default:
+		flipped = 0;
+		break;
+	}
+	
+	ev->ev_x = x;
+	ev->ev_y = y;
+	ev->ev_buttons = (buttons ^ flipped) | modflags;
+	ev->ev_flags = flipped;
+}
 #endif /* X11 */
 
 int
@@ -776,26 +914,25 @@ event_t *ev;
 	int new_x, new_y, xyset;
 
 	for (;;) {
-		do {
-			ev->ev_flags = 0;
-			xyset = FALSE;
-			if (Pushedback_event.ev_type != 0) {
-				*ev = Pushedback_event;
-				xyset = TRUE;
-				Pushedback_event.ev_type = 0;
-			}
-			else {
-				if (_wn_Replaying) {
-					if (wn_get_recorded_event(ev) == 0)
-						xyset = TRUE;
-					else
-						get_event(wn, ev, &xyset);
-				}
+		ev->ev_flags = 0;
+		xyset = FALSE;
+		
+		if (Pushedback_event.ev_type != 0) {
+			*ev = Pushedback_event;
+			xyset = TRUE;
+			Pushedback_event.ev_type = 0;
+		}
+		else {
+			if (_wn_Replaying) {
+				if (wn_get_recorded_event(ev) == 0)
+					xyset = TRUE;
 				else
 					get_event(wn, ev, &xyset);
 			}
-		} while ((mask & ev->ev_type) == 0);
-
+			else
+				get_event(wn, ev, &xyset);
+		}
+		
 		w = WN_TO_W(ev->ev_wn);
 		bw = w->w_bw;
 		if (!xyset) {
@@ -834,9 +971,13 @@ event_t *ev;
 		}
 		bw->bw_buttons = ev->ev_buttons;
 
+		if ((mask & ev->ev_type) == 0)
+			continue;
+
 		if (ev->ev_type == EV_WINDOW_RESIZED)
 			bw->bw_winch_event = EV_WINDOW_RESIZED;
-		else if (ev->ev_type == EV_WINDOW_EXPOSED && bw->bw_winch_event == 0)
+		else if (ev->ev_type == EV_WINDOW_EXPOSED &&
+			 bw->bw_winch_event == 0)
 			bw->bw_winch_event = EV_WINDOW_EXPOSED;
 		else if (ev->ev_type == EV_LOST_SELECTION)
 			bw->bw_had_lost_selection_event = TRUE;
@@ -947,9 +1088,9 @@ int *p_xpos, *p_ypos;
 }
 
 void
-wn_ungetpuck(wn,x,y,buttons)
+wn_ungetpuck(wn, x, y, buttons)
 int wn;
-int x,y,buttons;
+int x, y, buttons;
 {
 	event_t evbuf;
 
@@ -958,7 +1099,7 @@ int x,y,buttons;
 	evbuf.ev_type = EV_MOUSE_MOVED;
 	evbuf.ev_x = x;
 	evbuf.ev_y = y;
-	evbuf.ev_buttons = buttons & (B_ANY | B_SHIFT_KEY | B_CONTROL_KEY);
+	evbuf.ev_buttons = buttons & (B_ANY | B_ANY_MODIFIER);
 	evbuf.ev_wn = wn;
 	wn_pushback_event(&evbuf);
 }
@@ -987,6 +1128,33 @@ int unused_sig;
 	(*Abort_func)();
 }
 
+/*  This function is based on the one in Stevens.
+ */
+static Sigfunc
+safe_signal(sig, func)
+int sig;
+Sigfunc func;
+{
+	struct sigaction act, oact;
+
+	/*  God knows what the actual type of sa_handler will be on any
+	 *  given system ...
+	 */
+	*(Sigfunc *)&act.sa_handler = func;
+	sigemptyset(&act.sa_mask);
+
+#ifdef SA_RESTART
+	act.sa_flags = SA_RESTART;
+#else
+	act.sa_flags = 0;
+#endif
+
+	if (sigaction(sig, &act, &oact) < 0)
+		wn__panic("sigaction failed");
+
+	return (Sigfunc)oact.sa_handler;
+}
+
 wn_abort_func_t
 wn_set_abort_func(wn,func)
 int wn;
@@ -1012,14 +1180,34 @@ wn_abort_func_t func;
 
 	if (func != NULL) {
 		set_intr_event_mask(w);
-		old_io_func = signal(SIGIO, catch_sigio);
+		old_io_func = safe_signal(SIGIO, catch_sigio);
+#ifndef __hp9000s800
+#if defined(USE_STREAMS) && !defined(FASYNC)
+		if (ioctl(fd, I_SETSIG, S_RDNORM) != 0) {
+			perror("wn: ioctl I_SETSIG S_RDNORM");
+			wn__panic("I_SETSIG S_RDNORM botch");
+		}
+#else
 		if (fcntl(fd, F_SETFL, FASYNC) != 0)
 			wn__panic("FASYNC on botch");
+#endif
+#endif /* __hp9000s800 */
 	}
 	else {
-		(void) signal(SIGIO, old_io_func);
+#ifdef FASYNC			/* RGA linux merge */
+		(void) safe_signal(SIGIO, old_io_func);
+#ifndef __hp9000s800
+#ifdef USE_STREAMS
+		if (ioctl(fd, I_SETSIG, 0) != 0) {
+			perror("wn: ioctl I_SETSIG 0");
+			wn__panic("I_SETSIG 0 botch");
+		}
+#else
 		if (fcntl(fd, F_SETFL, 0) != 0)
 			wn__panic("FASYNC off botch");
+#endif
+#endif /* __hp9000s800 */
+#endif /* RGA linux merge */
 		set_normal_event_mask(w);
 	}
 
@@ -1060,3 +1248,60 @@ swin_t *w;
 	wn__setmode(w->w_wn);
 #endif
 }
+
+int
+wn_expose_or_press_queued(wn, md, p_x, p_y, exposed, resized)
+     int wn;
+     int md;
+     int *p_x;
+     int *p_y;
+     int *exposed;
+     int *resized;
+{
+#ifdef X11
+  XEvent x_event;
+  XMotionEvent ev;
+  swin_t *w;
+  event_t event;
+  int x, y, xyset, ret = 0;
+
+  if (_wn_Windowtab)
+  {
+    /* first send a dummy event so we are guaranteed not to block */
+    memset((char *)&ev, 0, sizeof(XMotionEvent));
+    ev.type = KeyPress;
+    ev.display = wn__Dpy;
+    ev.window = WN_TO_W(WN_STDWIN)->w_win;
+    XSendEvent(wn__Dpy, ev.window, False, 0L, (XEvent *)&ev); 
+    *exposed = *resized = 0;
+
+    for(;;)
+    {
+      XPeekEvent(wn__Dpy, &x_event);
+     if (x_event.type == ButtonPress)
+      {
+	w = WN_TO_W(wn);
+	x = x_event.xbutton.x - w->w_x_offs;
+	y = x_event.xbutton.y - w->w_y_offs;
+	if (Mgetmd(x, y, wn) == md)
+	{
+	  *p_x = x;
+	  *p_y = y;
+	  ret = 1;
+	}
+      }
+      /* read all events till we get the send event at the end */
+      get_event(wn, &event, &xyset);
+      if (event.ev_type == EV_WINDOW_EXPOSED)
+	*exposed = 1;
+      else
+      if (event.ev_type == EV_WINDOW_RESIZED)
+	*resized = 1;
+
+      if (x_event.xany.send_event)
+	return ret;
+    }
+  }
+#endif /* X11 */
+  return 0;
+}
