@@ -7,10 +7,10 @@
  *  unaltered copy of this message (including the copyright).
  */
 
-char wn_wn_init_c_sccsid[] = "@(#)wn_init.c	1.44 15/9/92 (UKC)";
+char wn_wn_init_c_sccsid[] = "@(#)wn_init.c	1.53 25 May 1995 (UKC)";
 
-#ifdef SVR4
-#include <sys/fcntl.h>
+#if defined(__svr4__) || defined(__SVR4)
+#define SVR4 1
 #endif
 
 #include <sys/types.h>
@@ -21,18 +21,18 @@ char wn_wn_init_c_sccsid[] = "@(#)wn_init.c	1.44 15/9/92 (UKC)";
 #include <signal.h>
 #include <string.h>
 #include <stdlib.h>
+#include <fcntl.h>
 
 #ifdef __STDC__
 #include <unistd.h>
 #endif
 
-#include <local/ukcprog.h>
-
 #include "wn.h"
 #include "wn_priv.h"
 
 #ifdef X11
 #include <X11/Xresource.h>
+#include <X11/Xatom.h>
 #endif
 
 #include "wn_bm.h"
@@ -46,30 +46,19 @@ char wn_wn_init_c_sccsid[] = "@(#)wn_init.c	1.44 15/9/92 (UKC)";
 #include "wn_color.h"
 #include "sccsdata.h"
 
-#if defined(clipper) || (defined(vax) && !defined(ultrix))
-#define NO_PUTENV
-#endif
-
-#ifdef NO_PUTENV
-static int my_putenv PROTO((const char *s));
-#define putenv	my_putenv
-#endif
-
-#ifndef SVR4
-/*  BUG: this shouldn't be here, it should be in a system header file.
- */
-int putenv PROTO((const char *s));
-#endif
+static int change_env_var PROTO((const char *s));
 
 #ifdef X11
 static const char *get_xdef PROTO((const char *name, const char *def));
-static int _wn_create_x_window PROTO((swin_t *w, const char *name, int x, int y, int width, int height, int border_width, const char *geometry));
+static int create_x_window PROTO((swin_t *w, const char *name,
+						const char *geometry));
 static void create_gc PROTO((void));
 static void set_colors PROTO((void));
 static Display *open_display PROTO((const char *displayname, const char **p_errmsg));
-static void parse_x11_geometry PROTO((XSizeHints *xsh, const char *geometry));
 static void get_def_color PROTO((const char *defaultname, xpixel_t def_pixel, int def_rgb_val, const char *argspec, XColor *color));
-static int _wn_create_x_window PROTO((swin_t *w, const char *name, int x, int y, int width, int height, int border_width, const char *geometry));
+static int   /* RCB: Changed from type void to match X spec */
+non_fatal_x_error_handler PROTO((Display *displayname, XErrorEvent *err));
+static void initialize_atoms PROTO((Window p_win));
 #endif
 
 #ifdef SUNVIEW
@@ -77,7 +66,6 @@ static int my_error_handler PROTO((int errnum, int winopnum));
 #endif
 
 static int add_wm_fd PROTO((int fd, char *errbuf));
-static void parse_x11_geometry PROTO(( XSizeHints *xsh, const char *geometry));
 
 #ifdef SUNVIEW
 static int (*Old_sun_error_handler)();
@@ -96,6 +84,9 @@ static bool App_suggests_xpos = FALSE, App_suggests_ypos = FALSE;
 static bool App_suggests_width = FALSE, App_suggests_height = FALSE;
 static int App_xpos, App_ypos, App_width, App_height;
 
+static int App_width_inc, App_height_inc, App_base_width, App_base_height;
+static bool App_suggests_resize_inc = FALSE;
+
 /*  Values for the foreground and background pixels.
  *  Initialised for mono displays - color ones overwrite them.
  */
@@ -114,6 +105,10 @@ int wn__Batching_enabled = TRUE;
 /*  Kludge for ICL guide - see _wn_dont_map_windows_on_create() and _wn_map_stdwin().
  */
 static int Map_windows_on_create = TRUE;
+static short atomic_world_initialized= FALSE;
+static Atom
+  atom_wm_protocols, atom_wm_delete_window, atom_wm_change_state;
+
 #endif /* X11 */
 
 /*  BUG: need a comment explaining why these aren't all ifdef'ed.
@@ -128,7 +123,7 @@ static const char *Appclass = Default_appclass;
 static int Want_to_start_as_icon = FALSE;
 
 #ifdef X11
-Display *wn__Dpy;
+Display *wn__Dpy = NULL;
 #ifdef X11
 Colormap _wn_Cmap;
 extern int _Xdebug;
@@ -143,12 +138,77 @@ int _wn_Autoraise = FALSE;
 const char *_wn_Icon_geometry = NULL;
 
 static bool Want_reverse_video = FALSE;
+static int   /* RCB: Changed from type void to match X spec */
+non_fatal_x_error_handler(d, err)
+     Display *d;
+     XErrorEvent *err;
+
+{
+  char str[80];
+
+  XGetErrorText(d, (int)err->error_code, str, 80);
+  fprintf(stderr, "UPS trapped non fatal X error: %s\n",str);
+  return 0;
+}
+
+/*
+ * Process quit on frame menu
+ */
+void process_client_message(event, frame)
+     XEvent *event;
+     Window frame;
+
+{
+  XClientMessageEvent *xclient	= &event->xclient;
+  Atom protocol;
+  
+  if( atomic_world_initialized 
+     && xclient->message_type == atom_wm_protocols)
+  {
+    protocol = event->xclient.data.l[0];
+    if( protocol == atom_wm_delete_window)
+    {
+      if (target_process_running())
+	target_stop();
+      re_set_exit_event_loop_flag();
+    }
+  }
+}
+
+/*
+ * Initialize atoms to catch quit on frame menu
+ */
+static void initialize_atoms(frame)
+     Window frame;
+     
+{
+  Atom atomic_interest[1];
+  atom_wm_change_state = XInternAtom (wn__Dpy, "WM_CHANGE_STATE", False);
+  atom_wm_protocols = XInternAtom(wn__Dpy, "WM_PROTOCOLS", FALSE);
+  
+  if(atom_wm_protocols != None)
+  {
+    atomic_world_initialized= TRUE;
+    atom_wm_delete_window = XInternAtom(wn__Dpy, "WM_DELETE_WINDOW", FALSE);
+    atomic_interest[0] = atom_wm_delete_window;
+    XChangeProperty(wn__Dpy, frame, atom_wm_protocols, XA_ATOM,
+		    32, PropModeReplace, (unsigned char*)atomic_interest, 1);
+  }
+}
+
 #endif /* X11 */
 
+void set_window_error_handler()
+{
+#ifdef X11
+  XSetErrorHandler(non_fatal_x_error_handler);
+#endif /* X11 */
+}
+
 const char *
 wn_version()
 {
-	return _wn_sccsdata[0];
+	return wn__sccsdata[0];
 }
 
 void
@@ -217,6 +277,12 @@ const char *name;
 #endif /* !X11 */
 }
 
+int
+wn_use_mono()
+{
+	return wn__Use_mono;
+}
+
 void
 wn_use_color_hint(use_color)
 int use_color;
@@ -285,7 +351,7 @@ const char **argv;
 	const char **src, **dst, **lim, *s;
 	int len;
 #ifdef X11
-	const char *font, *pos;
+	const char *font;
 #endif /* X11 */
 
 	if ((Appname = strrchr(*argv, '/')) != NULL && Appname[1] != '\0')
@@ -299,7 +365,7 @@ const char **argv;
 	 *  fish this out first.  We must also get any -name flag.
 	 */
 	for (src = dst; src < lim - 1; ++src) {
-		if (strlen(*src) > 1 && strncmp(*src, "-display",
+		if (strlen(*src) > (size_t)1 && strncmp(*src, "-display",
 						         strlen(*src)) == 0)
 			Default_display = *++src;
 		else if (strcmp(*src, "-name") == 0)
@@ -318,6 +384,14 @@ const char **argv;
 	if ((s = wn_get_default("ForceMono")) != NULL && strcmp(s, "yes") == 0)
 		wn__Use_mono = TRUE;
 
+	if ((s = wn_get_default("WantMenuColors")) != NULL &&
+	    strcmp(s, "yes") == 0 && wn__Use_mono == FALSE)
+		wn__Use_menu_colors = TRUE;
+
+	if ((s = wn_get_default("WantExtendedTextColors")) != NULL &&
+	    strcmp(s, "yes") == 0 && wn__Use_mono == FALSE)
+		wn__Use_extended_text_colors = TRUE;
+
 	/*  We allow auto raise to be set by an environment variable because
 	 *  passing arguments is hard for ICL guide.
 	 */
@@ -325,10 +399,17 @@ const char **argv;
 		_wn_Autoraise = TRUE;
 	else
 		_wn_Autoraise = wn_get_default("AutoRaise") != NULL;
+
+       wn_set_MultiClickTime();	/* RCB */
 #endif /* X11 */
 
 	if (getenv("KSTMONO") != NULL)
-		wn__Use_mono = TRUE;
+	{
+	  wn__Use_mono = TRUE;
+	  wn__Use_alloc_color = FALSE;
+	  wn__Use_menu_colors = FALSE;
+	  wn__Use_extended_text_colors = FALSE;
+	}
 	if (getenv("KSTCOLOR") != NULL || getenv("KSTCOLOUR") != NULL)
 		wn__Use_mono = FALSE;
 
@@ -348,13 +429,13 @@ const char **argv;
 			++src;
 		}
 #ifdef X11
-		else if (len > 1 && strncmp(s, "-display", len) == 0 &&
+		else if (len > 1 && strncmp(s, "-display", (size_t)len) == 0 &&
 								src[1] != NULL) {
 			_wn_add_to_unmunge_list(src, 2);
 			Default_display = src[1];
 			src += 2;
 		}
-		else if (len > 1 && strncmp(s, "-geometry", len) == 0 &&
+		else if (len > 1 && strncmp(s, "-geometry", (size_t)len) == 0 &&
 								src[1] != NULL) {
 			Stdwin_geometry = src[1];
 			src += 2;
@@ -368,10 +449,6 @@ const char **argv;
 			++src;
 		}
 #endif /* X11 */
-		else if ((pos = strchr(s, ':')) != NULL && isdigit(pos[1])) {
-			_wn_add_to_unmunge_list(src, 1);
-			Default_display = *src++;
-		}
 		else if (strcmp(s, "-fn") == 0 && src[1] != NULL) {
 			_wn_add_to_unmunge_list(src, 1);
 			_wn_Sysfont_file = src[1];
@@ -404,6 +481,9 @@ const char **argv;
 		else if (strcmp(s, "-mono") == 0) {
 			_wn_add_to_unmunge_list(src, 1);
 			wn__Use_mono = TRUE;
+			wn__Use_alloc_color = FALSE;
+			wn__Use_menu_colors = FALSE;
+			wn__Use_extended_text_colors = FALSE;
 			++src;
 		}
 		else if (strcmp(s, "-wn_debug") == 0) {
@@ -448,8 +528,15 @@ wn_open_stdwin()
 const char *
 _wn_reason()
 {
-	extern char *sys_errlist[];
-	extern int errno, sys_nerr;
+#ifndef __FreeBSD__
+#ifdef __linux__
+extern const char *const sys_errlist[];
+extern int sys_nerr;
+#else
+extern const char* sys_errlist[];
+extern int sys_nerr;
+#endif
+#endif
 
 	return (errno > 0 && errno < sys_nerr) ? sys_errlist[errno]
 					       : "unknown reason";
@@ -524,14 +611,26 @@ set_colors()
 	want_mono = wn__Use_mono || (v->class != PseudoColor &&
 						    v->class != DirectColor);
 
-	if (!want_mono) {
-		if (XAllocColorCells(wn__Dpy, _wn_Cmap, TRUE,
-						&planes, 1, &rw_pixel, 1) == 0) {
-			fprintf(stderr, "%s: Using monochrome as there are no free colormap entries\n", Appname);
-
-			want_mono = wn__Use_mono = TRUE;
-		}
-		
+	if (!want_mono && wn__Use_alloc_color == FALSE) {
+	  if (XAllocColorCells(wn__Dpy, _wn_Cmap, TRUE,
+			       &planes, 1, &rw_pixel, 1) == 0)
+	  {
+	    if (v->class == PseudoColor)
+	    {
+	      want_mono = TRUE;
+	      wn__Use_alloc_color = TRUE;
+	      wn__Use_alloc_color_for_highlight = TRUE;
+	    }
+	    else
+	    {
+	      fprintf(stderr, "%s: Using monochrome as there are no free colormap entries\n", Appname);
+	      
+	      want_mono = wn__Use_mono = TRUE;
+	      wn__Use_alloc_color = TRUE;
+	      wn__Use_menu_colors = FALSE;
+	      wn__Use_extended_text_colors = FALSE;
+	    }
+	  }
 	}
 
 	if (want_mono) {
@@ -539,8 +638,11 @@ set_colors()
 		 *  sent to me by Jur van der Burg <vdburg@utrtsc.enet.dec.com>.
 		 *  I an unable to test it as we have no displays of this
 		 *  type locally.
+		 *  RGA: added TrueColor clause from Craig Amey 
+		 *  <craiga@MetaSW.COM>
 		 */
-		if (v->class == StaticGray || v->class == GrayScale) {
+		if (v->class == StaticGray || v->class == GrayScale ||
+		    v->class == TrueColor) {
 			int i, j;
 
 			j = DisplayPlanes(wn__Dpy, DefaultScreen(wn__Dpy));
@@ -562,6 +664,11 @@ set_colors()
 			bg_color.pixel = Want_reverse_video ? black_pixel
 							    : white_pixel;
 
+		if (!wn__Use_mono)
+		{
+		  wn__Use_alloc_color = TRUE;
+		  wn__Use_alloc_color_for_highlight = TRUE;
+		}
 		_wn_Fg_pixel = fg_color.pixel;
 		_wn_Bg_pixel = bg_color.pixel;
 
@@ -629,10 +736,11 @@ const char **p_errmsg;
 			display_fd = ConnectionNumber(display);
 #endif /* X11 */
 
-			buf = wn__e_malloc(strlen(real_displayname) + sizeof(deq));
-			(void) strcpy(buf, deq);
-			(void) strcat(buf, real_displayname);
-			(void) putenv(buf);
+			buf = wn__e_malloc(strlen(real_displayname) +
+					   sizeof(deq));
+			strcpy(buf, deq);
+			strcat(buf, real_displayname);
+			change_env_var(buf);
 
 			if (add_wm_fd(display_fd, errbuf) != 0) {
 				XCloseDisplay(display);
@@ -652,21 +760,17 @@ add_wm_fd(fd, errbuf)
 int fd;
 char *errbuf;
 {
+#if defined(F_SETOWN) && !defined(SVR4)
 	if (fcntl(fd, F_SETFD, 1) != 0) {
-#ifdef SVR4
-		/*  Richard Almeida says F_SETOWN fails on almost all SVR4
-		 *  systems.
-		 */
-#else
 		if (errbuf != NULL) {
 			sprintf(errbuf,
 				"Can't set close-on-exec flag for display fd %d (%s)\n",
 				fd, _wn_reason());
 		}
 		return -1;
-#endif
 	}
 
+#ifdef FASYNC			/* RGA linux merge */
 	if (fcntl(fd, F_SETOWN, getpid()) != 0) {
 		if (errbuf != NULL) {
 			sprintf(errbuf,
@@ -674,34 +778,37 @@ char *errbuf;
 			 	fd, _wn_reason());
 		}
 		return -1;
+#endif /* RGA linux merge */
 	}
+#endif /* !SVR4 */
 
 	_wn_change_wn_fdmask(fd);
 	return 0;
 }
 
-#ifdef NO_PUTENV
-/*  Define putenv for machines the don't have it in the standard library.
+/*  This function has the same interface as putenv(), but the #ifdefs to
+ *  use putenv() only where it exists are more hassle than they are worth.
  */
 static int
-my_putenv(s)
+change_env_var(s)
 const char *s;
 {
-	int nlen;
+	size_t nlen;
 	const char *cptr;
-	const char **nenv, **eptr;
-	extern const char **environ;
+	char **nenv, **eptr;
+	extern char **environ;	  /* as specified by POSIX.1 section 2.6*/
 
 	/*  First see if there is an existing 'name=value' with the
 	 *  same name as s.
 	 */
 	for (cptr = s; *cptr != '=' && *cptr != '\0'; cptr++)
 		;
+	
 	if (*cptr == '=' && cptr > s) {
 		nlen = cptr - s + 1;
 		for (eptr = environ; *eptr != NULL; eptr++) {
 			if (strncmp(*eptr, s, nlen) == 0) {
-				*eptr = s;
+				*eptr = (char *)s;
 				return 0;
 			}
 		}
@@ -711,18 +818,22 @@ const char *s;
 	 */
 	for (eptr = environ; *eptr != NULL; eptr++)
 		;
-	nenv = (const char **) malloc((eptr - environ + 2) * sizeof(char *));
+	
+	nenv = (char **)malloc((eptr - environ + 2) * sizeof(char *));
 	if (nenv == NULL)
 		return -1;
+	
 	eptr = environ;
 	environ = nenv;
+	
 	while ((*nenv++ = *eptr++) != NULL)
 		;
-	*nenv = s;
+	
+	*nenv = (char *)s;
 	nenv[1] = NULL;
+	
 	return 0;
 }
-#endif /* NO_PUTENV */
 
 #ifdef X11
 static void
@@ -736,7 +847,7 @@ create_gc()
 
 	gcv.foreground = _wn_Fg_pixel;
 	gcv.background = _wn_Bg_pixel;
-	_wn_Gc = XCreateGC(wn__Dpy, rootwin, GCMASK, &gcv);
+	_wn_Gc = XCreateGC(wn__Dpy, rootwin, (unsigned)GCMASK, &gcv);
 }
 
 void
@@ -954,178 +1065,121 @@ swin_t *w;
 const char *name;
 int is_mainwin;
 {
-	XWindowAttributes infobuf;
-	Window bwin;
-	const char *geometry, *wins;
+	const char *geometry;
 
 	if (is_mainwin && Stdwin_geometry != NULL)
 		geometry = Stdwin_geometry;
 	else
 		geometry = get_xdef(name, "Geometry");
 
-	if (is_mainwin && (Want_blanket || wn_get_default("Blanket") != NULL)) {
-		if ((wins = getenv("WINDOWID")) != NULL) {
-			bwin = atoi(wins);
-			XGetWindowAttributes(wn__Dpy, bwin, &infobuf);
-		}
-	}
-	else {
-		memset((char *)&infobuf, '\0', sizeof(infobuf));
-	}
-
-	return _wn_create_x_window(w, name, infobuf.x, infobuf.y,
-						infobuf.width, infobuf.height,
-						infobuf.border_width, geometry);
-}
-#endif /* X11 */
-
-#ifdef X11
-/*  This version of parse_x11_geometry is from a version sent to
- *  me by der Mouse (<mouse@thunder.mcrcim.mcgill.edu>).
- *  I have reformatted and modified it a bit.
- */
-static void
-parse_x11_geometry(xsh, geometry)
-XSizeHints *xsh;
-const char *geometry;
-{
-	int x, y, flags;
-	unsigned w, h;
-
-	flags = XParseGeometry(geometry, &x, &y, &w, &h);
-
-	if ((flags & WidthValue) != 0)
-		xsh->width = w;
-	if ((flags & HeightValue) != 0)
-		xsh->height = h;
-	
-	/*  When XNegative, x still gets a negative number!  (Similarly y.)
-	 *
-	 *  The 4s are 2 * border_width, and border width is hardwired (!)
-	 *  to 2 in _wn_create_x_window() - der Mouse.
-	 */
-	if ((flags & XValue) != 0) {
-		if ((flags & XNegative) != 0) {
-			xsh->x = DisplayWidth(wn__Dpy, DefaultScreen(wn__Dpy))
-							   - xsh->width - 4 + x;
-		}
-		else {
-			xsh->x = x;
-     		}
-	}
-
-	if ((flags & YValue) != 0) {
-		if ((flags & YNegative) != 0) {
-			xsh->y = DisplayHeight(wn__Dpy,DefaultScreen(wn__Dpy)) -
-						     xsh->height - 4 + y;
-		}
-		else {
-			xsh->y = y;
-		}
-	}
-
-	if ((flags & (XValue | YValue)) != 0)
-		xsh->flags = (xsh->flags & ~PPosition) | USPosition;
-	
-	if ((flags & (WidthValue | HeightValue)) != 0)
-		xsh->flags = (xsh->flags & ~PSize) | USSize;
+	return create_x_window(w, name, geometry);
 }
 
 static int
-_wn_create_x_window(w, name, x, y, width, height, border_width, geometry)
+create_x_window(w, name, geometry)
 swin_t *w;
-const char *name;
-int x, y, width, height, border_width;
-const char *geometry;
+const char *name, *geometry;
 {
-	Window      win;		/* Window ID */
-	static XSizeHints  xsh;		/* Size hints for window manager */
+	XClassHint class_hint;
+	XWMHints xwmh;
+	Window win;
+	const char *bwstr;
+	XTextProperty nameprop/*, rga_nameprop*/;
+	XSizeHints sh;
+	int x, y;
+	unsigned flags, width, height, border_width;
 
-	/* Deal with providing the window with an initial position & size.
-	 * Fill out the XSizeHints struct to inform the window manager. See
-	 * Sections 9.1.6 & 10.3.
+	sh.flags = PPosition | PSize;
+	sh.x = App_suggests_xpos ? App_xpos : 0;
+	sh.y = App_suggests_ypos ? App_ypos : 0;
+	sh.width = App_suggests_width ? App_width : 650;
+	sh.height = App_suggests_height ? App_height : 550;
+
+	/*  Dubious - we want an application supplied xpos or ypos to place
+	 *  the window without prompting for position, and twm won't do this
+	 *  unless USPosition is set, so set it.  Must look at ICCCM sometime.
 	 */
-	memset((char *)&xsh, '\0', sizeof(xsh));
-	xsh.flags = PPosition | PSize;
-	if (width != 0) {
-		xsh.x = x;
-		xsh.y = y;
-		xsh.width = width;
-		xsh.height = height;
-	}
-	else {
-		xsh.x = App_suggests_xpos ? App_xpos : 0;
-		xsh.y = App_suggests_ypos ? App_ypos : 0;
-		xsh.width = App_suggests_width ? App_width : 650;
-		xsh.height = App_suggests_height ? App_height : 550;
-		border_width = 2;
-		if (geometry != NULL)
-			parse_x11_geometry(&xsh, geometry);
-		
-		/*  Dubious - we want an application supplied xpos or
-		 *  ypos to place the window without prompting for
-		 *  position, and twm won't do this unless USPosition
-		 *  is set, so set it.  Must look at ICCCM sometime.
-		 */
-		if (App_suggests_xpos || App_suggests_ypos)
-			xsh.flags = (xsh.flags & ~PPosition) | USPosition;
-	}
+	if (App_suggests_xpos || App_suggests_ypos)
+		sh.flags = (sh.flags & ~PPosition) | USPosition;
+
 	App_suggests_xpos = App_suggests_ypos = FALSE;
 	App_suggests_width = App_suggests_height = FALSE;
+	
+	border_width = 0;
+	if ((bwstr = wn_get_default("BorderWidth")) != NULL)
+		border_width = atoi(bwstr);
+	if (border_width == 0)
+		border_width = 2;
+
+	if (geometry != NULL)
+		flags = XParseGeometry(geometry, &x, &y, &width, &height);
+	else
+		flags = 0;
+
+	if ((flags & WidthValue) != 0) {
+		sh.width = width;
+		sh.flags |= USSize;
+	}
+	if ((flags & HeightValue) != 0) {
+		sh.height = height;
+		sh.flags |= USSize;
+	}
+
+	if (App_suggests_resize_inc) {
+		sh.base_width = App_base_width;
+		sh.base_height = App_base_height;
+		sh.width_inc = App_width_inc;
+		sh.height_inc = App_height_inc;
+
+		sh.flags |= PResizeInc | PBaseSize;
+
+		sh.width = sh.width * sh.width_inc + sh.base_width;
+		sh.height = sh.height * sh.height_inc + sh.base_height;
+		sh.min_width = sh.width_inc + sh.base_width;
+		sh.min_height = sh.height_inc + sh.base_height;
+
+		App_suggests_resize_inc = FALSE;
+	}
+
+
+	if (flags & XValue) {
+		if (flags & XNegative) {
+			x = DisplayWidth(wn__Dpy, DefaultScreen(wn__Dpy)) + x -
+					sh.width - 2 * border_width;
+		}
+		sh.x = x;
+		sh.flags |= USPosition;
+	}
+	if (flags & YValue) {
+		if (flags & YNegative) {
+			y = DisplayHeight(wn__Dpy, DefaultScreen(wn__Dpy)) + y -
+						sh.height - 2 * border_width;
+		}
+		sh.y = y;
+		sh.flags |= USPosition;
+	}
+
+	if (flags & XNegative) {
+		sh.win_gravity = (flags & YNegative) ? SouthEastGravity
+						     : NorthEastGravity;
+	}
+	else {
+		sh.win_gravity = (flags & YNegative) ? SouthWestGravity
+						     : NorthWestGravity;
+	}
+	sh.flags |= PWinGravity;
 
-	/*  Create the Window with the information in the XSizeHints, the
-	 *  border width,  and the border & background pixels. See Section 3.3.
-	 */
 	win = XCreateSimpleWindow(wn__Dpy,
 				  RootWindow(wn__Dpy, DefaultScreen(wn__Dpy)),
-				  xsh.x, xsh.y,
-				  (unsigned)xsh.width, (unsigned)xsh.height,
-				  (unsigned)border_width,
+				  sh.x, sh.y,
+				  (unsigned)sh.width, (unsigned)sh.height,
+				  border_width,
 				  (xpixel_t)_wn_Fg_pixel, (xpixel_t)_wn_Bg_pixel);
 
-	/*  Set the standard properties for the window managers. See Section 9.1.
-	 */
-	XSetStandardProperties(wn__Dpy, win, name, name, None,
-						(char **)&Appname, 1, &xsh);
-
-	XSelectInput(wn__Dpy, win, ExposureMask);
-	w->w_win = win;
-	return 0;
-}
-
-void
-_wn_map_X_window(w, iw)
-swin_t *w, *iw;
-{
-	XWMHints xwmh;
-	XEvent x_event;
-	XClassHint class_hint;
-
-	/*  BUG: The wn window creation stuff for X11 needs rewriting.
-	 *       The whole thing is a mess.
-	 */
 	xwmh.initial_state = Want_to_start_as_icon ? IconicState : NormalState;
 	xwmh.input = True;
 	xwmh.flags = InputHint | StateHint;
 
-	if (_wn_Icon_geometry != NULL) {
-		int res;
-		unsigned junk_width, junk_height;
-
-		res = XParseGeometry(_wn_Icon_geometry, &xwmh.icon_x, &xwmh.icon_y,
-						   &junk_width, &junk_height);
-		if ((res & (XValue | YValue)) == (XValue | YValue))
-			xwmh.flags |= IconPositionHint;
-	}
-
-	if (iw != NULL) {
-		xwmh.icon_window = iw->w_win;
-		xwmh.flags |= IconWindowHint;
-	}
-
-
-	XSetWMHints(wn__Dpy, w->w_win, &xwmh);
-
 	if (Appclass == Default_appclass && Appname != Default_appname) {
 		char *class;
 
@@ -1138,7 +1192,52 @@ swin_t *w, *iw;
 
 	class_hint.res_name = (char *)Appname;
 	class_hint.res_class = (char *)Appclass;
-	XSetClassHint(wn__Dpy, w->w_win, &class_hint);
+
+	if (XStringListToTextProperty((char **)&name, 1, &nameprop) == 0)
+		wn__panic("can't make text property for window name");
+
+	/*  Set the standard properties for the window managers.
+	 */
+
+	XSetWMProperties(wn__Dpy, win, &nameprop, &nameprop,
+				(char **)NULL, 0, &sh, &xwmh, &class_hint);
+
+	XSelectInput(wn__Dpy, win, ExposureMask);
+
+	w->w_win = win;
+	initialize_atoms(win);
+	return 0;
+}
+
+void
+_wn_map_X_window(w, iw)
+swin_t *w, *iw;
+{
+	XEvent x_event;
+	XWMHints xwmh;
+
+	/*  BUG: The wn window creation stuff for X11 needs rewriting.
+	 *       The whole thing is a mess.
+	 */
+
+	if (_wn_Icon_geometry != NULL) {
+		int res;
+		unsigned junk_width, junk_height;
+
+		res = XParseGeometry(_wn_Icon_geometry,
+						&xwmh.icon_x, &xwmh.icon_y,
+					        &junk_width, &junk_height);
+		if ((res & (XValue | YValue)) == (XValue | YValue))
+			xwmh.flags |= IconPositionHint;
+	}
+
+	if (iw != NULL) {
+		xwmh.icon_window = iw->w_win;
+		xwmh.flags |= IconWindowHint;
+
+		XSetWMHints(wn__Dpy, w->w_win, &xwmh);
+	}
+
 
 	if (Map_windows_on_create) {
 		XMapWindow(wn__Dpy, w->w_win);
@@ -1148,6 +1247,7 @@ swin_t *w, *iw;
 			} while (x_event.type != Expose);
 		}
 	}
+
 	wn__update_size(w);
 }
 #endif /* X11 */
@@ -1162,6 +1262,17 @@ int x, y;
 	App_ypos = y;
 }
 
+void
+wn_suggest_resize_inc(base_width, base_height, width_inc, height_inc)
+int base_width, base_height, width_inc, height_inc;
+{
+	App_base_width = base_width;
+	App_base_height = base_height;
+	App_width_inc = width_inc;
+	App_height_inc = height_inc;
+	App_suggests_resize_inc = TRUE;
+}
+
 void
 wn_suggest_window_size(width, height)
 int width, height;
@@ -1193,7 +1304,8 @@ int wn;
 	w = WN_TO_W(wn);
 
 	xwa.override_redirect = TRUE;
-	XChangeWindowAttributes(wn__Dpy, w->w_win, CWOverrideRedirect, &xwa);
+	XChangeWindowAttributes(wn__Dpy, w->w_win,
+					(unsigned)CWOverrideRedirect, &xwa);
 
 	win = Want_to_start_as_icon ? w->w_bw->bw_assocw->w_win : w->w_win;
 	w->w_bw->bw_is_mapped = !Want_to_start_as_icon;
@@ -1203,3 +1315,52 @@ int wn;
 	wn__update_size(w);
 #endif
 }
+
+/* Function to determine if a TrueColor visual is being used */
+/* RGA: from Craig Amey <craiga@MetaSW.COM> */
+
+int wn_is_truecolor()
+{
+  Visual *v;
+
+  v = DefaultVisual(wn__Dpy, DefaultScreen(wn__Dpy));
+  return (v->class == TrueColor);
+}
+
+/* Function used to allocate shared colors in TrueColor mode */
+/* RGA: from Craig Amey <craiga@MetaSW.COM> */
+
+int wn_alloc_color(color, colorname)
+color_t *color;
+const char *colorname;
+{
+#ifdef X11
+  XColor xcolor;
+
+  xcolor.red = color->co_red;
+  xcolor.green = color->co_green;
+  xcolor.blue = color->co_blue;
+
+  if (XAllocColor(wn__Dpy, _wn_Cmap, &xcolor) == 0)
+  {
+    fprintf(stderr, "ups: Failed to allocate color `%s'\n", colorname);
+    color->co_pixel = WN_FG;
+    wn__Use_alloc_color = FALSE;
+    wn__Use_alloc_color_for_highlight = TRUE; /* note */
+    wn__Use_extended_text_colors = FALSE;
+    wn__Use_menu_colors = FALSE;
+    return 0;
+  }
+  else
+  {
+    color->co_pixel = xcolor.pixel;
+    wn_use_alloc_color(1);	/* set */
+    wn_use_alloc_color_for_highlight(1);	/* set */
+    return 1;
+  }
+#else
+  color->co_pixel = WN_FG;
+  return 1;
+#endif
+}
+
