@@ -7,7 +7,7 @@
  *  unaltered copy of this message (including the copyright).
  */
 
-char ups_ci_types_c_sccsid[] = "@(#)ci_types.c	1.13 20/5/92 (UKC)";
+char ups_ci_types_c_sccsid[] = "@(#)ci_types.c	1.18 04 Jun 1995 (UKC)";
 
 #include <string.h>
 
@@ -27,6 +27,11 @@ static const char *nametype_to_string PROTO((nametype_t nametype));
 static bool params_same PROTO((funcret_t *fr1, funcret_t *fr2));
 static char *param_types PROTO((type_t *type));
 static const char *qualifiers_to_string PROTO((qualifiers_t qualifiers));
+static typecode_t flags_to_typecode PROTO((unsigned typeflags));
+static unsigned typecode_to_flag PROTO((typecode_t typecode,
+							bool alltypes_ok));
+static const char *typeflag_to_name PROTO((unsigned flag));
+static void complain_about_bad_type PROTO((unsigned flags));
 
 #define CHAR_SIZE	sizeof(char)
 #define SHORT_SIZE	sizeof(short)
@@ -43,16 +48,34 @@ static const char *qualifiers_to_string PROTO((qualifiers_t qualifiers));
 #define LONG_ALIGN	LONG_SIZE
 #define POINTER_ALIGN	POINTER_SIZE
 #define FLOAT_ALIGN	FLOAT_SIZE
-#define DOUBLE_ALIGN	FLOAT_SIZE	/* BUG: dubious */
+#define DOUBLE_ALIGN	FLOAT_SIZE	/* BUG: wrong (esp. on SPARC) */
 
 #define WORD_ALIGN	LONG_ALIGN
 #define MAX_ALIGN	DOUBLE_ALIGN
 
+#define TF_TYPE_DONE	0x0001
+#define TF_VOID		0x0002
+#define TF_CHAR		0x0004
+#define TF_SHORT	0x0008
+#define TF_INT		0x0010
+#define TF_LONG		0x0020
+#define TF_FLOAT	0x0040
+#define TF_DOUBLE	0x0080
+#define TF_SIGNED	0x0100
+#define TF_UNSIGNED	0x0200
+#define TF_ENUM		0x0400
+#define TF_STRUCT	0x0800
+#define TF_UNION	0x1000
+
+#define TF_MAX_MASK	TF_UNION
+
+#define TF_ENUM_OR_AGGR	(TF_ENUM | TF_STRUCT | TF_UNION)
+
 int
 ci_type_alignment(type)
 type_t *type;
 {
-	switch(type->ty_code) {
+	switch(ci_effective_typecode(type->ty_code)) {
 	case TY_U_ENUM:
 	case TY_ENUM:
 	case DT_PTR_TO:
@@ -68,6 +91,7 @@ type_t *type;
 	case TY_INT:
 	case TY_UINT:
 	case TY_BITFIELD:
+	case TY_BOOLEAN:
 		return INT_ALIGN;
 	case TY_SHORT:
 	case TY_USHORT:
@@ -85,6 +109,32 @@ type_t *type;
 	}
 }
 
+optype_t
+ci_typecode_to_cvt_op(typecode)
+typecode_t typecode;
+{
+	switch (ci_effective_typecode(typecode)) {
+		case TY_CHAR:	return OP_CVT_TO_CHAR;
+		case TY_UCHAR:	return OP_CVT_TO_UCHAR;
+		case TY_SHORT:	return OP_CVT_TO_SHORT;
+		case TY_USHORT:	return OP_CVT_TO_USHORT;
+		case TY_INT:	return OP_CVT_TO_INT;
+		case TY_U_ENUM:	return OP_CVT_TO_INT;
+		case TY_ENUM:	return OP_CVT_TO_INT;
+		case TY_UINT:	return OP_CVT_TO_UINT;
+		case TY_LONG:	return OP_CVT_TO_LONG;
+		case TY_ULONG:	return OP_CVT_TO_ULONG;
+		case TY_FLOAT:	return OP_CVT_TO_FLOAT;
+		case TY_VOID:	return OP_CVT_TO_VOID;
+		case TY_DOUBLE:	return OP_CVT_TO_DOUBLE;
+		case DT_PTR_TO:	return OP_CVT_TO_PTR;
+		case TY_BOOLEAN:return OP_CVT_TO_INT;
+		default:
+			ci_panic("bad type in pc");
+			return OP_LAST; /* to satisfy gcc */
+	}
+}
+
 taddr_t
 ci_align_addr(addr, alignment)
 taddr_t addr;
@@ -107,69 +157,254 @@ type_t *type;
 	return ci_align_addr(addr, alignment);
 }
 
-void
-ci_fix_signed_and_unsigned(declaration)
-declaration_t *declaration;
+static void
+complain_about_bad_type(flags)
+unsigned flags;
 {
-	if (declaration->dn_basetype != NULL) {
-		if (declaration->dn_basetype->ty_code == TY_UNSIGNED)
-			declaration->dn_basetype = ci_code_to_type(TY_UINT);
-		else if (declaration->dn_basetype->ty_code == TY_SIGNED)
-			declaration->dn_basetype = ci_code_to_type(TY_INT);
+	unsigned mask;
+	int i, count;
+	const char *typeword[20];
+	char *s;
+
+	if ((flags & (TF_SIGNED | TF_INT)) == (TF_SIGNED | TF_INT))
+		flags &= ~TF_INT;
+
+	count = 0;
+	for (mask = 1; mask <= TF_MAX_MASK; mask <<= 1) {
+		if (flags & mask)
+			typeword[count++] = typeflag_to_name(mask);
+	}
+
+	if (count == 2) {
+		diagf(ET_ERROR, (lexinfo_t *)NULL,
+		      "Both `%s' and `%s' given in declaration",
+						typeword[0], typeword[1]);
+		return;
+	}
+
+	if (count < 3)
+		panic("count botch in cabt");
+	
+	s = strf("`%s'", typeword[0]);
+
+	for (i = 1; i < count - 1; ++i) {
+		char *new;
+
+		new = strf("%s, `%s'", s, typeword[i]);
+		free(s);
+		s = new;
+	}
+	
+	diagf(ET_ERROR, (lexinfo_t *)NULL,
+	      "%s and `%s' given in declaration", s, typeword[i]);
+	free(s);
+}
+
+type_t *
+ci_get_dn_basetype(dn)
+declaration_t *dn;
+{
+	if ((dn->dn_typeflags & TF_TYPE_DONE) == 0) {
+		typecode_t typecode;
+
+		switch (typecode = flags_to_typecode(dn->dn_typeflags)) {
+		case TY_NOTYPE:
+			complain_about_bad_type(dn->dn_typeflags);
+			dn->dn__basetype = NULL;
+			break;
+		case TY_STRUCT:
+			break;
+		default:
+			dn->dn__basetype = ci_code_to_type(typecode);
+		}
+
+		dn->dn_typeflags |= TF_TYPE_DONE;
+	}
+
+	return dn->dn__basetype;
+}
+	
+static const char *
+typeflag_to_name(flag)
+unsigned flag;
+{
+	switch (flag) {
+		case TF_VOID:		return "void";
+		case TF_CHAR:		return "char";
+		case TF_SHORT:		return "short";
+		case TF_INT:		return "int";
+		case TF_LONG:		return "long";
+		case TF_FLOAT:		return "float";
+		case TF_DOUBLE:		return "double";
+		case TF_SIGNED:		return "signed";
+		case TF_UNSIGNED:	return "unsigned";
+		case TF_ENUM:		return "enum";
+		case TF_STRUCT:		return "struct";
+		case TF_UNION:		return "union";
+		default:		panic("bad flag in ttn");
+	}
+
+	return NULL;	/* to satisfy gcc */
+}
+
+static typecode_t
+flags_to_typecode(typeflags)
+unsigned typeflags;
+{
+	switch (ci_effective_typecode(typeflags)) {
+	case TF_VOID:
+		return TY_VOID;
+
+	case TF_CHAR:
+	case TF_SIGNED | TF_CHAR:
+		/* BUG: `char' and `signed char' are distinct types.
+		 */
+		return TY_CHAR;
+
+	case TF_UNSIGNED | TF_CHAR:
+		return TY_UCHAR;
+	
+	case TF_SHORT:
+	case TF_SIGNED | TF_SHORT:
+	case TF_SHORT | TF_INT:
+	case TF_SIGNED | TF_SHORT | TF_INT:
+		return TY_SHORT;
+	
+	case TF_UNSIGNED | TF_SHORT:
+	case TF_UNSIGNED | TF_SHORT | TF_INT:
+		return TY_USHORT;
+
+	case TF_INT:
+	case TF_SIGNED:
+	case TY_BOOLEAN:
+	case TF_SIGNED | TF_INT:
+		return TY_INT;
+	case 0:
+		diagf(ET_WARNING, (lexinfo_t *)NULL, "Type defaults to int");
+		return TY_INT;
+	
+	case TF_UNSIGNED:
+	case TF_UNSIGNED | TF_INT:
+		return TY_UINT;
+	
+	case TF_LONG:
+	case TF_SIGNED | TF_LONG:
+	case TF_LONG | TF_INT:
+	case TF_SIGNED | TF_LONG | TF_INT:
+		return TY_LONG;
+
+	case TF_UNSIGNED | TF_LONG:
+	case TF_UNSIGNED | TF_LONG | TF_INT:
+		return TY_ULONG;
+	
+	case TF_FLOAT:
+		return TY_FLOAT;
+
+	case TF_DOUBLE:
+	case TF_LONG | TF_DOUBLE:
+		/* BUG: `long double' and `double' are distinct types.
+		 */
+		return TY_DOUBLE;
+	
+	case TF_ENUM:
+	case TF_STRUCT:
+	case TF_UNION:
+		return TY_STRUCT;	/* see ci_get_dn_basetype() */
+	
+	default:
+		return TY_NOTYPE;
 	}
 }
 
-void
-ci_add_type_specifier(declaration, oldtype)
-declaration_t *declaration;
-type_t *oldtype;
+static unsigned
+typecode_to_flag(typecode, alltypes_ok)
+typecode_t typecode;
+bool alltypes_ok;
 {
-	typecode_t oldcode, newcode;
-	type_t *newtype, *type;
-
-	newtype = declaration->dn_basetype;
-	oldcode = oldtype->ty_code;
-	newcode = newtype->ty_code;
-
-	if ((oldcode == TY_SHORT || oldcode == TY_LONG) && newcode == TY_INT)
-		type = oldtype;
-	else if (oldcode == TY_LONG && newcode == TY_DOUBLE)
-		type = newtype;
-	else if (oldcode == TY_UNSIGNED || oldcode == TY_SIGNED) {
-		typecode_t ucode;
-
-		switch (newcode) {
-			case TY_CHAR:	ucode = TY_UCHAR;	break;
-			case TY_SHORT:	ucode = TY_USHORT;	break;
-			case TY_INT:	ucode = TY_UINT;	break;
-			case TY_LONG:	ucode = TY_ULONG;	break;
-			default:	ucode = TY_NOTYPE;	break;
-		}
-		if (ucode != TY_NOTYPE) {
-			if (oldcode == TY_UNSIGNED)
-				type = ci_code_to_type(ucode);
-			else
-				type = newtype;
-		}
-		else
-			type = NULL;
+	switch (ci_effective_typecode(typecode)) {
+	case TY_VOID:
+		return TF_VOID;
+	case TY_CHAR:
+		return TF_CHAR;
+	case TY_SHORT:
+		return TF_SHORT;
+	case TY_BOOLEAN:
+	case TY_INT:
+		return TF_INT;
+	case TY_LONG:
+		return TF_LONG;
+	case TY_FLOAT:
+		return TF_FLOAT;
+	case TY_DOUBLE:
+		return TF_DOUBLE;
+	case TY_SIGNED:
+		return TF_SIGNED;
+	case TY_UNSIGNED:
+		return TF_UNSIGNED;
+	case TY_STRUCT:
+	case TY_U_STRUCT:
+		return TF_STRUCT;
+	case TY_UNION:
+	case TY_U_UNION:
+		return TF_UNION;
+	case TY_ENUM:
+	case TY_U_ENUM:
+		return TF_ENUM;
+	default:
+		if (alltypes_ok)
+			return TF_TYPE_DONE;
+		panic("bad typecode in ctt");	
+		return 0;	/* to satisfy gcc */
 	}
-	else
-		type = NULL;
+}
 
-	if (type != NULL)
-		declaration->dn_basetype = type;
+declaration_t *
+ci_make_declaration(class, basetype, qualifiers, declarators)
+class_t class;
+type_t *basetype;
+qualifiers_t qualifiers;
+declarator_t *declarators;
+{
+	unsigned typeflags;
+	declaration_t *dn = NEW(declaration_t);
+
+	if (basetype == NULL) {
+		basetype = ci_code_to_type(TY_INT);
+		typeflags = 0;
+	}
 	else {
-		char *oldtypestr, *newtypestr;
-
-		oldtypestr = ci_type_to_english(oldtype, FALSE);
-		newtypestr = ci_type_to_english(newtype, FALSE);
-		diagf(ET_ERROR, (lexinfo_t *)NULL,
-		      "`%s' and `%s' in the same declaration",
-							oldtypestr, newtypestr);
-		free(oldtypestr);
-		free(newtypestr);
+		typeflags = typecode_to_flag(basetype->ty_code, TRUE);
 	}
+
+	dn->dn_class = class;
+	dn->dn__basetype = basetype;
+	dn->dn_typeflags = typeflags;
+	dn->dn_dup_typeflags = 0;
+	dn->dn_qualifiers = qualifiers;
+	dn->dn_declarators = declarators;
+	dn->dn_next = NULL;
+
+	return dn;
+}
+
+void
+ci_add_type_specifier(dn, modtype)
+declaration_t *dn;
+type_t *modtype;
+{
+	unsigned flag;
+
+	flag = typecode_to_flag(modtype->ty_code, FALSE);
+
+	if ((dn->dn_typeflags & flag) != 0 &&
+					(dn->dn_dup_typeflags & flag) == 0) {
+		diagf(ET_WARNING, (lexinfo_t *)NULL,
+		      "Duplicate use of `%s' in declaration",
+					typeflag_to_name(flag));
+		dn->dn_dup_typeflags |= flag;
+	}
+
+	dn->dn_typeflags |= flag;
 }
 
 /*  Return the size in bytes of type.
@@ -180,19 +415,20 @@ lexinfo_t *lx;
 type_t *type;
 {
 	if (type->ty_size == -1) {
-		taddr_t size;
+		long size;
 
-		switch(type->ty_code) {
+		switch(ci_effective_typecode(type->ty_code)) {
 		case DT_PTR_TO:
 			size = POINTER_SIZE;
 			break;
 		case DT_ARRAY_OF:
-			size = type->ty_dim->di_high *
+			size = (type->ty_dim->di_high + 1) *
 						ci_typesize(lx, type->ty_base);
 			break;
 		case DT_FUNC_RETURNING:
-			diagf(ET_ERROR, lx, "Sizeof <function> undefined");
-			size = -1;
+/* RGA			diagf(ET_ERROR, lx, "Sizeof <function> undefined");
+			size = -1;*/
+			size = 0;
 			break;
 		case TY_VOID:
 			diagf(ET_ERROR, lx, "Sizeof void undefined");
@@ -200,9 +436,10 @@ type_t *type;
 			break;
 		case TY_U_STRUCT:
 		case TY_U_UNION:
-			diagf(ET_ERROR, lx,
+/*	RGA		diagf(ET_ERROR, lx,
 			      "Sizeof <%s> undefined", ci_basetype_name(type));
-			size = -1;
+			size = -1;*/
+			size = 0;
 			break;
 		case TY_STRUCT:
 		case TY_UNION:
@@ -219,6 +456,7 @@ type_t *type;
 		case TY_INT:
 		case TY_UINT:
 		case TY_BITFIELD:
+		case TY_BOOLEAN:
 			size = INT_SIZE;
 			break;
 		case TY_SHORT:
@@ -236,9 +474,10 @@ type_t *type;
 			size = DOUBLE_SIZE;
 			break;
 		default:
-			ci_panic("unknown typecode in ts");
-			size = -1; /* to satisfy gcc */
-		}
+/*			ci_panic("unknown typecode in ts");*/
+/*			size = -1;*/ /* to satisfy gcc */
+			size = 0;
+	}
 		type->ty_size = size;
 	}
 	return type->ty_size;
@@ -333,12 +572,14 @@ type_t *t1, *t2;
 				if (!ci_types_same(d1->di_type, d2->di_type))
 					return FALSE;
 
-				if (!d1->di_ldynamic && !d2->di_ldynamic &&
-							d1->di_low != d2->di_low)
+				if (d1->di_ltype == DT_CONSTANT &&
+				    d2->di_ltype == DT_CONSTANT &&
+				    d1->di_low != d2->di_low)
 					return FALSE;
 				
-				if (!d1->di_hdynamic && !d2->di_hdynamic &&
-						    d1->di_high != d2->di_high)
+				if (d1->di_htype == DT_CONSTANT &&
+				    d2->di_htype == DT_CONSTANT &&
+				    d1->di_high != d2->di_high)
 					return FALSE;
 				
 				return TRUE;
@@ -351,6 +592,7 @@ type_t *t1, *t2;
 				return FALSE;
 			break;
 		case TY_VOID:
+		case TY_BOOLEAN:
 		case TY_CHAR:	case TY_UCHAR:
 		case TY_SHORT:	case TY_USHORT:
 		case TY_INT:	case TY_UINT:
@@ -389,6 +631,7 @@ typecode_t typecode;
 	case TY_ENUM:
 	case TY_FLOAT:
 	case TY_DOUBLE:
+	case TY_BOOLEAN:
 		return TRUE;
 	case TY_UCHAR:
 	case TY_USHORT:
@@ -417,6 +660,7 @@ typecode_t typecode;
 	case TY_LONG:
 	case TY_ULONG:
 	case TY_BITFIELD:
+	case TY_BOOLEAN:
 		return TRUE;
 	default:
 		return FALSE;
@@ -469,6 +713,20 @@ type_t *type;
 	return buf;
 }
 
+/*  Rewrite any parameters of type "array of T" to "pointer to T"
+ */
+void
+ci_rewrite_func_params(fr)
+funcret_t *fr;
+{
+	var_t *v;
+
+	for (v = fr->fr_params; v != NULL; v = v->va_next) {
+		if (v->va_type->ty_code == DT_ARRAY_OF)
+			v->va_type = ci_make_pointer(v->va_type->ty_base, 0);
+	}
+}
+
 static char *
 param_types(type)
 type_t *type;
@@ -540,13 +798,14 @@ bool resolve_typedefs;
 			free(cptr);
 			break;
 		case DT_ARRAY_OF:
-			if (type->ty_dim->di_hdynamic)
+			if (type->ty_dim->di_htype != DT_CONSTANT) {
 				asizestr = "";
+			}
 			else{
 				static char buf[30];
 
-				strnf(buf, sizeof(buf), "%d", 
-							type->ty_dim->di_high);
+				strnf(buf, sizeof(buf), "%ld", 
+				      type->ty_dim->di_high + 1);
 				asizestr = buf;
 			}
 			new = strf(last_was_ptr ? "(%s)[%s]" : "%s[%s]",
@@ -589,7 +848,7 @@ bool resolve_typedefs;
 {
 	bool quitloop;
 	const char *space;
-	char *str, *cptr, *params_str;
+	char *str;
 
 	str = strsave("");
 	space = "";		/* no space the first time round the loop */
@@ -603,6 +862,9 @@ bool resolve_typedefs;
 			quitloop = TRUE;
 		}
 		else {
+			char *cptr;
+			const char *params_str;
+
 			switch (type->ty_code) {
 			case DT_PTR_TO:
 				new = strf("%s%s%spointer to", str, space,
@@ -616,13 +878,13 @@ bool resolve_typedefs;
 				free(cptr);
 				break;
 			case DT_ARRAY_OF:
-				if (type->ty_dim->di_hdynamic)
+				if (type->ty_dim->di_htype != DT_CONSTANT)
 					new = strf("%s%sarray[(unspecified size)] of",
 							str, space);
 				else
-					new = strf("%s%sarray[%d] of",
-							str, space,
-							type->ty_dim->di_high);
+					new = strf("%s%sarray[%ld] of",
+						   str, space,
+						   type->ty_dim->di_high + 1);
 				break;
 			default:
 				new = strf("%s%s%s", str, space,
