@@ -4,7 +4,7 @@
 static MENU MM1 = {
 	0x40,
 	0,
-	"clear",
+	"Clear",
 	99,
 	-1,
 	-1,
@@ -13,6 +13,8 @@ static MENU MM1 = {
 	0,
 	0,
 	0,
+	0,
+	0,
 	NULL,
 	NULL,
 	NULL,
@@ -22,7 +24,7 @@ static MENU MM1 = {
 static MENU MM2 = {
 	0x40,
 	0,
-	"backwards",
+	"Backwards",
 	98,
 	61,
 	1,
@@ -31,6 +33,8 @@ static MENU MM2 = {
 	0,
 	0,
 	0,
+	0,
+	0,
 	NULL,
 	NULL,
 	NULL,
@@ -40,7 +44,7 @@ static MENU MM2 = {
 static MENU MM3 = {
 	0x40,
 	0,
-	"forwards",
+	"Forwards",
 	102,
 	184,
 	1,
@@ -49,6 +53,8 @@ static MENU MM3 = {
 	0,
 	0,
 	0,
+	0,
+	0,
 	NULL,
 	NULL,
 	NULL,
@@ -67,6 +73,8 @@ static MENU MM4 = {
 	0,
 	0,
 	0,
+	0,
+	0,
 	NULL,
 	&MM2,
 	&MM3,
@@ -76,7 +84,7 @@ static MENU MM4 = {
 static MENU MM5 = {
 	0x260,
 	0,
-	"search",
+	"Search",
 	120,
 	125,
 	-1,
@@ -85,6 +93,8 @@ static MENU MM5 = {
 	0,
 	0,
 	0,
+	0,
+	0,
 	NULL,
 	&MM4,
 	NULL,
@@ -103,6 +113,8 @@ static MENU MM6 = {
 	0,
 	0,
 	0,
+	0,
+	0,
 	NULL,
 	&MM1,
 	&MM5,
@@ -112,8 +124,8 @@ static MENU MM6 = {
 static MENU MM7 = {
 	0x40,
 	0,
-	"up",
-	117,
+	"Load",
+	108,
 	252,
 	-1,
 	379,
@@ -121,6 +133,8 @@ static MENU MM7 = {
 	0,
 	0,
 	0,
+	0,
+	0,
 	NULL,
 	NULL,
 	NULL,
@@ -130,8 +144,8 @@ static MENU MM7 = {
 static MENU MM8 = {
 	0x40,
 	0,
-	"down",
-	100,
+	"Save",
+	115,
 	379,
 	-1,
 	506,
@@ -139,6 +153,8 @@ static MENU MM8 = {
 	0,
 	0,
 	0,
+	0,
+	0,
 	NULL,
 	NULL,
 	NULL,
@@ -157,6 +173,8 @@ static MENU MM9 = {
 	0,
 	0,
 	0,
+	0,
+	0,
 	NULL,
 	&MM7,
 	&MM8,
@@ -175,6 +193,8 @@ MENU out_men = {
 	0,
 	0,
 	0,
+	0,
+	0,
 	NULL,
 	&MM6,
 	&MM9,
