@@ -4,7 +4,7 @@
 MENU tgt_next_men = {
 	0x440,
 	0,
-	"next",
+	"Next",
 	110,
 	-1,
 	-1,
@@ -13,6 +13,8 @@ MENU tgt_next_men = {
 	0,
 	-30584,
 	0,
+	0,
+	0,
 	NULL,
 	NULL,
 	NULL,
