@@ -2,16 +2,18 @@
 #include <local/menu3.h>
 
 MENU tgt_step_men = {
-	0x440,
-	0,
-	"step",
+	0x460,
+	144,
+	"Step",
 	115,
 	-1,
 	-1,
 	100,
 	86,
 	0,
-	-30584,
+	2,
+	0,
+	0,
 	0,
 	NULL,
 	NULL,
