@@ -7,7 +7,7 @@
  *  unaltered copy of this message (including the copyright).
  */
 
-char ups_obj_stack_c_sccsid[] = "@(#)obj_stack.c	1.12 18/9/92 (UKC)";
+char ups_obj_stack_c_sccsid[] = "@(#)obj_stack.c	1.21 20 Jun 1995 (UKC)";
 
 #include <mtrprog/ifdefs.h>
 
@@ -23,29 +23,31 @@ char ups_obj_stack_c_sccsid[] = "@(#)obj_stack.c	1.12 18/9/92 (UKC)";
 
 #include <local/wn.h>
 #include <local/obj/obj.h>
+#include <local/obj/o_sel.h>
+#include <local/obj/o_mkobj.h>
+
 #include "objtypes.h"
 #include "ups.h"
 #include "symtab.h"
-
+#include "ao_syms.h"
+#include "st.h"
+#include "target.h"
 #include "obj_stack.h"
-#include "proc.h"
-#include "core.h"
 #include "data.h"
-#include "preamble.h"
 #include "ui.h"
 #include "va.h"
 #include "obj_signal.h"
 #include "obj_misc.h"
-#include "stack.h"
 #include "menudata.h"
 #include "debug.h"
 #include "tdr.h"
+#include "state.h"
 
 /*  Arguments block for label_if_var_addr.
  */
 typedef struct label_if_addr_argsst {
 	func_t *li_func;
-	stack_t *li_stk;
+	Stack *li_stk;
 	taddr_t li_addr;
 	char *li_buf;
 } label_if_var_addr_args_t;
@@ -53,7 +55,7 @@ typedef struct label_if_addr_argsst {
 /*  A block of a function in a stack.
  */
 typedef struct stackblockst {
-	stack_t *sb_stk;
+	Stack *sb_stk;
 	block_t *sb_block;
 } stackblock_t;
 
@@ -62,140 +64,30 @@ static void check_stack PROTO((void));
 #endif
 static char *badfuncname PROTO((taddr_t addr));
 static void line_draw PROTO((struct drawst *dets));
-static objid_t add_sig_markers PROTO((int lastsig));
+static objid_t add_sig_markers PROTO((target_t *xp));
 static void add_func_vars PROTO((objid_t par));
 static int remove_stack_object_if_not_a_func PROTO((objid_t par, fval_t unused_arg));
-static void set_func_fields PROTO((func_t *f, int lnum, fval_t *fields));
+static void set_func_fields PROTO((func_t *f, fil_t *fil,
+				   int lnum, fval_t *fields));
 static void expand_block PROTO((objid_t par, block_t *block));
 static void label_if_var_addr PROTO((var_t *v, char *args));
-static stack_t *stackpos_to_stk PROTO((stackpos_t pos));
-static stack_t *obj_to_stk PROTO((objid_t obj));
+static Stack *stackpos_to_stk PROTO((stackpos_t pos));
 static void pathname_draw PROTO((struct drawst *dets));
+static bool is_non_empty_block PROTO((block_t *par));
+static Stack *obj_to_stack PROTO((objid_t obj));
+static void check_for_f90_rterr PROTO((target_t *xp, Stack *stk));
 
 /*  Pointers to the outermost and innermost frame in the list of stack frames.
  *  NULL if there is no stack trace.
  */
-static stack_t *Outer_stack = NULL;
-static stack_t *Inner_stack = NULL;
+static Stack *Outer_stack = NULL;
+static Stack *Inner_stack = NULL;
 
 /*  Pointer to the stack frame of the "current function".  This is the
  *  function that next and step refer to, and where local variables
  *  are initially searched for.
  */
-static stack_t *Current_stack = NULL;
-
-/*  Attempt to get the argv array of the target from the stack so
- *  we can put the arguments in the command line.  If we succeed,
- *  put as many arguments as will fit in command_line, which is
- *  taken to be command_line_size bytes long, and return 0.
- *
- *  If we fail, return -1 (with no error message).
- *
- *  FRAGILE CODE
- *  
- *  This function assumes the following stack layout, starting from
- *  the top of the stack:
- *
- *	Zero or more words with the value zero
- *	The actual bytes of the argument and environment strings
- *	A zero word
- *	Zero or more pointers to environment strings
- *	A zero word
- *	One or more pointers to argument strings
- *
- *  We get confused if the argument bytes contain a zero word - this
- *  can happen if a command is invoked with four consecutive zero length
- *  arguments.
- *
- *  It's not fatal for this routine to fail - all that happens is that
- *  no arguments are displayed on the target line.
- */
-int
-get_command_line_from_stack(command_line, command_line_size)
-char *command_line;
-int command_line_size;
-{
-	static char dotdotdot[] = " ...";
-	enum { INITIAL, IN_STRINGS, IN_ENVP, IN_ARGV, FINAL } state;
-	char argbuf[512];
-	char *lptr, *lim;
-	int i, argc;
-	taddr_t *argv;
-	taddr_t ibuf[128];
-	taddr_t *iptr, word, sp, maxsp;
-
-	argc = 0; /* gcc doesn't understand state machines */
-	sp = maxsp = get_max_stack_addr();
-	iptr = ibuf;
-
-	for (state = INITIAL; state != FINAL; ) {
-
-		/*  Safety limit - if the stack is bad we don't
-		 *  want to read the entire core.
-		 */
-		if (maxsp - sp > 30000)
-			return -1;
-
-		if (iptr == ibuf) {
-			if (dread(sp - sizeof(ibuf), (char *)ibuf, sizeof(ibuf)) != 0)
-				return -1;
-			iptr = ibuf + sizeof(ibuf) / sizeof(ibuf[0]);
-		}
-
-		sp -= 4;
-		word = *--iptr;
-
-		switch(state) {
-		case INITIAL:
-			if (word != 0)
-				state = IN_STRINGS;
-			break;
-		case IN_STRINGS:
-			if (word == 0)
-				state = IN_ENVP;
-			break;
-		case IN_ENVP:
-			if (word == 0) {
-				state = IN_ARGV;
-				argc = 0;
-			}
-			break;
-		case IN_ARGV:
-			if (word == argc)
-				state = FINAL;
-			else
-				++argc;
-			break;
-		default:
-			panic("bad state in gclafs");
-		}
-	}
-
-	/*  At this point the start of the argument vector is at sp.
-	 */
-	if (argc == 0)
-		return -1;
-	argv = (taddr_t *)e_malloc(argc * sizeof(taddr_t));
-	if (dread(sp + 4, (char *)argv, argc * sizeof(taddr_t)) != 0)
-		return -1;
-	
-	lptr = command_line;
-	lim = command_line + command_line_size - sizeof(dotdotdot);
-
-	for (i = 0; i < argc; ++i) {
-		if (dgets(argv[i], argbuf, sizeof(argbuf)) != 0)
-			return -1;
-		if (lptr + strlen(argbuf) >= lim) {
-			(void) strcpy(lptr, dotdotdot);
-			break;
-		}
-		if (i != 0)
-			*lptr++ = ' ';
-		(void) strcpy(lptr, argbuf);
-		lptr += strlen(argbuf);
-	}
-	return 0;
-}
+static Stack *Current_stack = NULL;
 
 taddr_t
 get_outer_fp()
@@ -233,7 +125,7 @@ taddr_t addr;
 bool *p_is_labelled;
 {
 	static char buf[100];
-	stack_t *stk, *above_stk, *below_stk;
+	Stack *stk, *above_stk, *below_stk;
 	label_if_var_addr_args_t libuf;
 	func_t *f;
 
@@ -244,11 +136,11 @@ bool *p_is_labelled;
 		f = stk->stk_func;
 		if (stk->stk_fp == addr) {
 			(void) sprintf(buf, "[%4x %s:fp]",
-							addr & 0xffff, f->fu_name);
+					     (int)addr & 0xffff, f->fu_demangled_name);
 			return buf;
 		}
 		if (stk->stk_ap == addr) {
-			(void) sprintf(buf, "%s:<ap>", f->fu_name);
+			(void) sprintf(buf, "%s:<ap>", f->fu_demangled_name);
 			*p_is_labelled = TRUE;
 			return buf;
 		}
@@ -272,15 +164,16 @@ bool *p_is_labelled;
 	}
 
 	if (above_stk != NULL && above_stk->stk_fp - addr < 256)
-		(void) sprintf(buf, "%4x+%02x ", above_stk->stk_fp & 0xffff,
-					        above_stk->stk_fp - addr);
+		(void) sprintf(buf, "%4x+%02x ",
+			       (int)above_stk->stk_fp & 0xffff,
+			       (int)(above_stk->stk_fp - addr));
 	else
 		(void) sprintf(buf, "%8s", "");
 
 	if (below_stk != NULL && addr - below_stk->stk_fp < 256)
 		(void) sprintf(buf + strlen(buf), "%4x-%02x",
-						below_stk->stk_fp & 0xffff,
-						addr - below_stk->stk_fp);
+					       (int)below_stk->stk_fp & 0xffff,
+					       (int)(addr - below_stk->stk_fp));
 	else
 		(void) sprintf(buf, "%7s", "");
 
@@ -295,8 +188,9 @@ get_current_func(p_func, p_fp, p_pc, p_adjusted_pc)
 func_t **p_func;
 taddr_t *p_fp, *p_pc, *p_adjusted_pc;
 {
-	if (Current_stack == NULL)
+	if (Current_stack == NULL) {
 		*p_func = NULL;
+	}
 	else {
 		*p_func = Current_stack->stk_func;
 		*p_pc = Current_stack->stk_pc;
@@ -314,13 +208,19 @@ taddr_t *p_fp, *p_pc, *p_adjusted_pc;
 	}
 }
 
+char *
+get_current_stack_data()
+{
+	return (Current_stack != NULL) ? Current_stack->stk_data : NULL;
+}
+
 static char *
 badfuncname(addr)
 taddr_t addr;
 {
 	char buf[50];
 
-	(void) sprintf(buf, "<bad text address 0x%x>", addr);
+	(void) sprintf(buf, "<bad text address 0x%lx>", addr);
 	return strsave(buf);
 }
 
@@ -330,29 +230,35 @@ static void
 line_draw(dets)
 register struct drawst *dets;
 {
-	wn_draw_line(dets->dr_wn,
-			dets->dr_x,
-			dets->dr_y+dets->dr_depth/2,
-			dets->dr_x + dets->dr_width,
-			dets->dr_y+dets->dr_depth/2,
-			WN_FG);
+  if (dets->selected)
+    wn_draw_line(dets->dr_wn,
+		 dets->dr_x,
+		 dets->dr_y+dets->dr_depth/2,
+		 dets->dr_x + dets->dr_width,
+		 dets->dr_y+dets->dr_depth/2,
+		 WN_BG);
+  else
+    wn_draw_line(dets->dr_wn,
+		 dets->dr_x,
+		 dets->dr_y+dets->dr_depth/2,
+		 dets->dr_x + dets->dr_width,
+		 dets->dr_y+dets->dr_depth/2,
+		 WN_FG);
 }
 
 fdef_t Fsig_fdefs[] = {
-	'l', line_draw, no_edit, NULL,
-	'\0', NULL, NULL, NULL,
+	{ 'l', line_draw, no_edit, NULL	},
+	{ '\0', NULL, NULL, NULL	},
 };
 
-const char Fsig_format[] = "%*100l%15cs%*100l\n";
+const char Fsig_format[] = "%*80l %20cs %*80l\n";
 #define FN_FSIG_SIGNAME 0
 
-int
-fsig_dumpobj(arg, code, level)
-char *arg;
+char *
+fsig_format_obj(code)
 objid_t code;
-int level;
 {
-	return td_outf(level, "%s", get_field_value(code, FN_FSIG_SIGNAME));
+	return strf("%s", get_field_value(code, FN_FSIG_SIGNAME));
 }
 
 void
@@ -363,35 +269,36 @@ objid_t obj;
 }
 
 static objid_t
-add_sig_markers(lastsig)
-int lastsig;
+add_sig_markers(xp)
+target_t *xp;
 {
-	stack_t *stk, *last;
+	Stack *stk, *last;
+	int lastsig;
 
 	last = NULL;
 	for (stk = Outer_stack; stk != NULL; stk = stk->stk_inner) {
 		last = stk;
 		if (stk->stk_siginfo != NULL) {
-			siginfo_t *si;
+			Siginfo *si;
 			const char *tag;
 
 			si = stk->stk_siginfo;
-			new_object((objid_t)si, OT_FSIG, (objid_t)stk, OBJ_BEFORE);
-			if (si->si_signo == -1)
-				tag = "<unknown signal number>";
-			else
-				tag = signame(si->si_signo);
+			new_object((objid_t)si, OT_FSIG, (objid_t)stk,
+								OBJ_BEFORE);
+			tag = xp_get_signal_tag(xp, si->si_signo);
 			set_field_value((objid_t)si, FN_FSIG_SIGNAME,
 							(fval_t)strsave(tag));
 		}
 	}
 
+	lastsig = xp_get_lastsig(xp);
 	if (last != NULL && lastsig != 0) {
 		const char *tag;
 
-		tag = strsave(signame(lastsig));
+		tag = xp_get_signal_tag(xp, lastsig);
 		new_object((objid_t)tag, OT_FSIG, (objid_t)last, OBJ_AFTER);
-		set_field_value((objid_t)tag, FN_FSIG_SIGNAME, (fval_t)tag);
+		set_field_value((objid_t)tag, FN_FSIG_SIGNAME,
+							(fval_t)strsave(tag));
 
 		return (objid_t)tag;
 	}
@@ -399,6 +306,7 @@ int lastsig;
 	return NULL;
 }
 
+
 /*  Delete stack objects other then functions.  This function is passed
  *  to visit_objects() to clean the old stack before building a new one.
  */
@@ -452,19 +360,34 @@ struct drawst *dets;
 			++s;
 			x += font->ft_width;
 		}
-		wn_ttext(dets->dr_wn, dots, x, dets->dr_y, dets->dr_fg, dets->dr_bg);
-		x += (sizeof(dots) - 1) * font->ft_width;
+/* AMC enhancements */
+	  if (dets->selected)
+	    wn_wd_ttext(dets->dr_wn, dots, x, dets->dr_y,
+			dets->dr_width, dets->dr_bg, dets->dr_fg);
+	  else
+	    wn_wd_ttext(dets->dr_wn, dots, x, dets->dr_y,
+			dets->dr_width, dets->dr_fg, dets->dr_bg);
+	  x += (sizeof(dots) - 1) * font->ft_width;
 	}
 
-	wn_ttext(dets->dr_wn, s, x, dets->dr_y, dets->dr_fg, dets->dr_bg);
+/* AMC enhancements */
+	if (dets->selected)
+	  wn_wd_ttext(dets->dr_wn, s, x, dets->dr_y,dets->dr_width, dets->dr_bg, dets->dr_fg);
+	else
+	  wn_wd_ttext(dets->dr_wn, s, x, dets->dr_y,dets->dr_width, dets->dr_fg, dets->dr_bg);
 }
 
 fdef_t Func_fdefs[] = {
-	'p', pathname_draw, no_edit, NULL,
-	'\0', NULL, NULL, NULL,
+	{ 'p', pathname_draw, no_edit, NULL	},
+	{ '\0', NULL, NULL, NULL		},
 };
 
-const char Func_format[] = "%[-]32cp%25cp%[-]10cs\n";
+/*const char Func_format[] = "%[-]32cp%25cp%[-]10cs\n";*/
+/* AMC enhancements */
+/*   Important to allow at least 4 chars in pathname so insertion of
+**   ... in pathname_draw does not back up to before start of string!
+*/
+const char Func_format[] = "%[-]:4c:50cp%25cp%[-]10cs  \n";
 
 #define FN_FUNC_NAME	0
 #define FN_FUNC_FILE	1
@@ -472,24 +395,29 @@ const char Func_format[] = "%[-]32cp%25cp%[-]10cs\n";
 #define FN_FUNC_LAST	3
 
 static void
-set_func_fields(f, lnum, fields)
+set_func_fields(f, fil, lnum, fields)
 func_t *f;
+fil_t *fil;
 int lnum;
 fval_t *fields;
 {
-	fields[FN_FUNC_NAME] = (fval_t)f->fu_name;
+	fields[FN_FUNC_NAME] = (fval_t)f->fu_demangled_name; /* RGA */
 
-	/*  The test for lnum==0 is really just to make the test scripts
+	/*  The test for lnum==0 etc is really just to make the test scripts
 	 *  easier.  Without this frig you get stack entries like
 	 *  `abort.s:0'.  Doing this is easier than putting {?xxx|yyy}
 	 *  in all the scripts.
 	 */
-	if ((f->fu_flags & FU_NOSYM) != 0 || lnum == 0) {
+	if ((f->fu_flags & FU_NOSYM) != 0) {
 		fields[FN_FUNC_FILE] = (fval_t)"(no symbols)";
 		fields[FN_FUNC_LNUM] = (fval_t)strsave("");
 	}
+	else if (lnum == 0) {
+		fields[FN_FUNC_FILE] = (fval_t)"(no line numbers)";
+		fields[FN_FUNC_LNUM] = (fval_t)strsave("");
+	}
 	else {
-		fields[FN_FUNC_FILE] = (fval_t)f->fu_fil->fi_name;
+		fields[FN_FUNC_FILE] = (fval_t)fil->fi_name_only;
 		fields[FN_FUNC_LNUM] = (fval_t)strf(":%d", lnum);
 	}
 
@@ -502,31 +430,29 @@ const char Badfunc_format[] = "%[-]30cs\n";
 /*  Add function objects to the display tree.
  */
 objid_t
-add_functions(proc, par)
-proc_t proc;
+add_functions(xp, par)
+target_t *xp;
 objid_t par;
 {
-	stack_t *stk, *old_current_stack, *source_stk, *next_stk, *old_stk;
-	stack_t *prev_stk;
+	Stack *stk, *old_current_stack, *source_stk, *next_stk, *old_stk;
+	Stack *first_new_stk;
+	Stack *prev_stk;
 	objid_t outer, obj, next, newcode, obj_to_make_visible;
-	bool still_matching, old_current_stack_was_innermost;
+	bool old_current_stack_was_innermost;
 	fval_t fields[FN_FUNC_LAST + 1];
 
-	if (Dummyfunc.fu_preamble_id == NULL) {
-		static preamble_t prbuf;
-
-		Dummyfunc.fu_preamble_id = (preamble_id_t)&prbuf;
-	}
-
 	source_stk = NULL;
 
 	visit_objects(par, OBJ_CHILDREN, remove_stack_object_if_not_a_func,
 						(fval_t)0, (objid_t *)NULL);
 
 	old_stk = Outer_stack;
-	Outer_stack = build_stack_trace(proc);
+	
+	Outer_stack = xp_get_stack_trace(xp);
+	check_for_f90_rterr(xp, Outer_stack);
+	
 	outer = NULL;
-	still_matching = TRUE;
+	first_new_stk = NULL;
 	old_current_stack = Current_stack;
 	old_current_stack_was_innermost = Current_stack == NULL ||
 					  Current_stack->stk_inner == NULL;
@@ -543,8 +469,7 @@ objid_t par;
 	if (Current_stack != NULL) {
 		for (stk = Current_stack->stk_inner; stk != NULL;
 							 stk = stk->stk_inner) {
-			if (stk->stk_lnum != 0 &&
-					     stk->stk_func->fu_fil != NULL) {
+			if (stk->stk_lnum != 0 && stk->stk_fil != NULL) {
 				old_current_stack_was_innermost = FALSE;
 				break;
 			}
@@ -553,28 +478,32 @@ objid_t par;
 	
 	for (stk = Outer_stack; stk != NULL; stk = next_stk) {
 		next_stk = stk->stk_inner;
-		set_func_fields(stk->stk_func, stk->stk_lnum, fields);
+		set_func_fields(stk->stk_func, stk->stk_fil, stk->stk_lnum,
+				fields);
 		newcode = NULL;
-		if (still_matching) {
-			if (old_stk == NULL || old_stk->stk_func != stk->stk_func)
-				still_matching = FALSE;
+		
+		if (first_new_stk == NULL) {
+			if (old_stk == NULL ||
+			    old_stk->stk_func != stk->stk_func ||
+			    (old_stk->stk_func->fu_flags & FU_BAD) != 0) {
+				first_new_stk = stk;
+			}
 			else {
 				*old_stk = *stk;
+				stk->stk_siginfo = NULL;
 
 				if (outer == NULL)
 					Outer_stack = old_stk;
 				else
-					((stack_t *)outer)->stk_inner = old_stk;
+					((Stack *)outer)->stk_inner = old_stk;
 
-				newcode = (objid_t) old_stk;
+				newcode = (objid_t)old_stk;
+				
 				if (old_stk == old_current_stack)
 					Current_stack = old_stk;
-				old_stk = (stack_t *)get_code(newcode, OBJ_NEXT);
-#ifdef CHECK_STACK
-				if (stk->stk_isfree)
-					panic("freeing free stk");
-#endif
-				free_stk(stk);
+				old_stk = (Stack *)get_code(newcode, OBJ_NEXT);
+				
+				destroy_stk(stk);
 			}
 		}
 
@@ -582,19 +511,20 @@ objid_t par;
 			free((char *)get_field_value(newcode, FN_FUNC_LNUM));
 		}
 		else {
-			newcode = (objid_t) stk;
+			newcode = (objid_t)stk;
+			
 			if (outer == NULL)
 				new_object(newcode, OT_FUNC, par, OBJ_CHILD);
 			else
 				new_object(newcode, OT_FUNC, outer, OBJ_AFTER);
 		}
 
-		stk = (stack_t *) newcode;
+		stk = (Stack *) newcode;
 
-		if (stk->stk_lnum != 0 && stk->stk_func->fu_fil != NULL)
+		if (stk->stk_lnum != 0 && stk->stk_fil != NULL)
 			source_stk = stk;
 
-		if (stk->stk_func == &Dummyfunc) {
+		if ((stk->stk_func->fu_flags & FU_BAD) != 0) {
 			change_type(newcode, OT_BADFUNC);
 			set_field_value(newcode, FN_BADFUNC_NAME,
 						badfuncname(stk->stk_pc));
@@ -607,14 +537,18 @@ objid_t par;
 		outer = newcode;
 	}
 
-	obj_to_make_visible = add_sig_markers(proc_get_lastsig(proc));
+	obj_to_make_visible = add_sig_markers(xp);
 
 	if (Current_stack == NULL || old_current_stack_was_innermost)
 		Current_stack = source_stk;
-	if (source_stk != NULL)
-		highlight_source(source_stk->stk_func->fu_fil, source_stk->stk_lnum);
-	else
-		highlight_source((fil_t *)NULL, 0);
+
+	if (source_stk != NULL) {
+		set_and_display_stopline(source_stk->stk_fil,
+					 source_stk->stk_lnum);
+	}
+	else {
+		set_and_display_stopline((fil_t *)NULL, 0);
+	}
 
 	/*  Remove any objects remaining from the old stack trace.
 	 */
@@ -627,11 +561,11 @@ objid_t par;
 	/*  If the stack went bad, add an object to indicate that.
 	 */
 	if (Outer_stack == NULL || Outer_stack->stk_bad) {
-		char badstackbuf[50];
+		static char badstackbuf[50]; /* RGA make static for purify */
 
-		(void) sprintf(badstackbuf, "<bad frame pointer 0x%x>", 
+		(void) sprintf(badstackbuf, "<bad frame pointer 0x%lx>", 
 				(Outer_stack != NULL) ? Outer_stack->stk_fp
-						      : proc_getreg(proc, REG_FP));
+						      : xp_getreg(xp, UPSREG_FP));
 		new_object((objid_t)badstackbuf, OT_BADFRAME, par, OBJ_CHILD);
 		set_field_value((objid_t)badstackbuf, 0, strsave(badstackbuf));
 	}
@@ -645,13 +579,68 @@ objid_t par;
 	}
 	Inner_stack = prev_stk;
 
+	for (stk = first_new_stk; stk != NULL; stk = stk->stk_inner)
+		restore_stack_entry_display(stk, FALSE);
+
 	return obj_to_make_visible;
 }
 
+/*  Try to patch up the stack trace in the face of the way the epcf90
+ *  compiler handles run time diagnostics.  These are handled by jumping
+ *  to some epilogue code which calls f_rterr() to report the error.
+ *  Because of the initial jump we would end up reporting the wrong
+ *  line (usually the end statement) as the location of the error.
+ *  epcf90 (when runtime checks are switched on) stores the current line
+ *  number at -24(fp) on a SPARC (this comes from looking at object files,
+ *  not from any documentation), so we look there for the line number.
+ *
+ *  BUG: what about other f90 compilers and other architectures?
+ */
+static void
+check_for_f90_rterr(xp, stk)
+target_t *xp;
+Stack *stk;
+{
+	Stack *last;
+	int lnum;
+	lno_t *ln;
+	taddr_t addr;
+
+	last = NULL;
+	
+	for (; stk != NULL; stk = stk->stk_inner) {
+		if (last != NULL && !last->stk_bad &&
+		    last->stk_func->fu_language == LANG_F90 &&
+		    strcmp(stk->stk_func->fu_name, "f_rterr") == 0)
+			break;
+		last = stk;
+	}
+
+	if (stk == NULL)
+		return;
+
+	/*  Are we at the last line of the function?
+	 */
+	for (ln = FU_LNOS(last->stk_func); ln != NULL; ln = ln->ln_next)
+		if (ln->ln_num > last->stk_lnum)
+			return;
+
+	if (xp_read_data(xp, last->stk_fp - 0x18,
+			 (char *)&lnum, sizeof(lnum)) != 0)
+		return;
+	
+	addr = lnum_and_fil_to_addr(last->stk_func, last->stk_fil, lnum);
+
+	if (addr != 0) {
+		last->stk_lnum = lnum;
+		last->stk_pc = addr;
+	}
+}
+
 void
 update_local_variable_values()
 {
-	stack_t *stk;
+	Stack *stk;
 
 	/*  Refresh the variable values for stack frames which were
 	 *  there before.
@@ -660,13 +649,11 @@ update_local_variable_values()
 		update_vars_of((objid_t)stk, FALSE);
 }
 
-int
-badstack_dumpobj(arg, code, level)
-char *arg;
+char *
+badstack_format_obj(code)
 objid_t code;
-int level;
 {
-	return td_outf(level, "%s", get_field_value(code, 0));
+	return strf("%s", get_field_value(code, 0));
 }
 
 void
@@ -681,7 +668,9 @@ delete_functions(par)
 objid_t par;
 {
 	remove_object(par, OBJ_DESCENDENTS);
+	
 	Outer_stack = NULL;
+	Inner_stack = NULL;
 }
 
 const char *
@@ -689,7 +678,7 @@ func_getobjname(obj)
 objid_t obj;
 {
 	static char *last = NULL;
-	stack_t *stk;
+	Stack *stk;
 	const char *filename, *funcname, *lnumstr;
 	int level;
 
@@ -701,7 +690,7 @@ objid_t obj;
 		free(last);
 
 	level = 0;
-	for (stk = (stack_t *)obj; stk != NULL; stk = stk->stk_outer)
+	for (stk = (Stack *)obj; stk != NULL; stk = stk->stk_outer)
 		++level;
 
 	last = strf("%d-%s:%s%s", level, filename, funcname, lnumstr);
@@ -713,22 +702,16 @@ void
 free_func(obj)
 objid_t obj;
 {
-	stack_t *stk;
+	Stack *stk;
 
-	stk = (stack_t *) obj;
+	stk = (Stack *) obj;
 	free((char *)get_field_value(obj, FN_FUNC_LNUM));
-#ifdef CHECK_STACK
-	if (stk->stk_isfree)
-		panic("freeing free stk");
-#endif
-	free_stk(stk);
+	destroy_stk(stk);
 }
 
-int
-func_dumpobj(arg, code, level)
-char *arg;
+char *
+func_format_obj(code)
 objid_t code;
-int level;
 {
 	const char *funcname, *filename, *lnumstr;
 	
@@ -736,7 +719,15 @@ int level;
 	filename = get_field_value(code, FN_FUNC_FILE);
 	lnumstr = get_field_value(code, FN_FUNC_LNUM);
 
-	return td_outf(level, "%-20s%25s%-10s", funcname, filename, lnumstr);
+	return strf("%-20s%25s%-10s", funcname, filename, lnumstr);
+}
+
+int
+pre_do_func(command, p_arg)
+int command;
+char **p_arg;
+{
+	return check_single_expr(command == MR_ADD_EXPRESSION);
 }
 
 /*  Process the return from a function line menu.
@@ -744,42 +735,137 @@ int level;
  *  Tell the editor where we are.
  */
 void
-do_func(par, command)
+do_func(par, command, arg)
 objid_t par;
 int command;
+char *arg;
 {
-	struct filst *fil;
-	stack_t *stk;
+	fil_t *fil;
+	Stack *stk;
 
-	stk = (stack_t *)par;
+	stk = (Stack *)par;
 
 	switch(command) {
-	case MR_ADD_VARS:
-		add_func_vars(par);
+	case MR_EXPAND_COMPLETELY:
+	case MR_EXPAND_LIKE_BEFORE:
+		if (command == MR_EXPAND_COMPLETELY ||
+		    !restore_stack_entry_display(stk, TRUE)) {
+			add_func_vars(par);
+		}
 		break;
-	case MR_HIDE_VARS:
+		
+	case MR_COLLAPSE:
+		update_stack_entry(stk);
 		remove_object(par, OBJ_DESCENDENTS);
+		note_stack_entry_collapsed(stk);
 		break;
+		
 	case MR_DISPLAY_SOURCE:
-		fil = stk->stk_func->fu_fil;
+		fil = stk->stk_fil;
+		
 		if (fil == NULL) {
 			errf("`%s' was not compiled with the -g flag",
-							stk->stk_func->fu_name);
+			     stk->stk_func->fu_demangled_name);
 			return;
 		}
-		if (highlight_source(fil, stk->stk_lnum) == 0)
+		
+		if (set_and_display_stopline(fil, stk->stk_lnum))
 			Current_stack = stk;
+		
 		break;
+		
 	case MR_ADD_EXPRESSION:
-		get_fi_vars(stk->stk_func->fu_fil);	/* ensure globals loaded */
-		add_expr_object(par, FU_BLOCKS(stk->stk_func),
-						stk->stk_func->fu_language);
+		st_get_fi_vars(stk->stk_fil);	/* ensure globals loaded */
+		add_expr_object(par, FU_BLOCKS(stk->stk_func), "",
+				stk->stk_func->fu_language, OBJ_LAST_CHILD);
+		break;
+	case MR_SHOW_PATH:
+		fil = stk->stk_fil;
+		if (!show_source_path_assumed(fil, FALSE))
+		  errf("\b%s %s", "library", stk->stk_func->fu_symtab->st_path);
+		break;
+	case MR_SHOW_SOURCE_PATH:
+		fil = stk->stk_fil;
+		if (!show_source_path_used(fil, FALSE))
+		  errf("\b%s %s", "library", stk->stk_func->fu_symtab->st_path);
+		break;
+	case MR_REMATCH_SOURCE_FILE:
+		fil = stk->stk_fil;
+		if (fil->fi_name)
+		  open_source_file(fil, TRUE, FALSE);
+		show_source(fil, 1);
+		break;
+	case MR_RELOAD_SOURCE_FILE:
+		fil = stk->stk_fil;
+		if (fil->fi_name)
+		  open_source_file(fil, FALSE, TRUE);
+		show_source(fil, 1);
+		break;
+	case MR_FILE_DATES:
+		fil = stk->stk_fil;
+		show_file_and_target_dates(fil, stk->stk_func->fu_symtab);
 		break;
 	default:
 		panic("bad cmd in df");
 	}
 }
 
+void
+maybe_show_stack_entry_source(obj)
+objid_t obj;
+{
+	Stack *stk;
+
+	stk = (Stack *)obj;
+
+	if (stk->stk_fil == NULL) {
+		errf("\bNo source file known for function %s",
+		     stk->stk_func->fu_demangled_name);
+	}
+	else {
+		do_func(obj, MR_DISPLAY_SOURCE, (char *)NULL);
+	}
+}
+		
+void
+change_displayed_stack_level(up)
+bool up;
+{
+	Stack *stk;
+	const char *withsource;
+
+	stk = Current_stack;
+	
+	if (stk == NULL) {
+		errf("No stack entry currently displayed");
+		return;
+	}
+
+	withsource = "";
+	
+	for (;;) {
+		stk = up ? stk->stk_outer : stk->stk_inner;
+
+		if (stk == NULL) {
+			errf("Already looking at %smost stack frame%s",
+			     up ? "outer" : "inner", withsource);
+			return;
+		}
+
+		if (stk->stk_func->fu_fil != NULL)
+			break;
+
+		withsource = " with source";
+	}
+
+	if (set_and_display_stopline(stk->stk_fil, stk->stk_lnum)) {
+		Current_stack = stk;
+		clear_selection();
+		select_object((objid_t)stk, TRUE, OBJ_SELF);
+		ensure_visible((objid_t)stk);
+	}
+}
+
 #define FN_BLOCK_NAME 0
 const char Block_format[] = "%*50l%15cs%*50l\n";
 
@@ -808,15 +894,15 @@ block_t *block;
 
 /*  Get the stack from an OT_FUNC or OT_BLOCK object.
  */
-static stack_t *
-obj_to_stk(obj)
+static Stack *
+obj_to_stack(obj)
 objid_t obj;
 {
-	stack_t *stk;
+	Stack *stk;
 
 	switch(get_object_type(obj)) {
 	case OT_FUNC:
-		stk = (stack_t *)obj;
+		stk = (Stack *)obj;
 		break;
 	case OT_BLOCK:
 		stk = ((stackblock_t *)obj)->sb_stk;
@@ -828,6 +914,17 @@ objid_t obj;
 	return stk;
 }
 
+block_t *
+obj_to_block(obj)
+objid_t obj;
+{
+	if (get_object_type(obj) == OT_BLOCK)
+		return ((stackblock_t *)obj)->sb_block;
+
+	panic("bad objtype in otb");
+	return NULL;		/* to satisfy gcc */
+}
+
 objid_t
 add_block_object(par, block)
 objid_t par;
@@ -838,7 +935,7 @@ block_t *block;
 	char buf[100];
 
 	sb = (stackblock_t *)e_malloc(sizeof(stackblock_t));
-	sb->sb_stk = obj_to_stk(par);
+	sb->sb_stk = obj_to_stack(par);
 	sb->sb_block = block;
 
 	/*  Add the block before any other blocks, but after any variables.
@@ -904,11 +1001,9 @@ objid_t obj;
 	return last;
 }
 
-int
-block_dumpobj(arg, code, level)
-char *arg;
+char *
+block_format_obj(code)
 objid_t code;
-int level;
 {
 	const char *name;
 
@@ -917,38 +1012,52 @@ int level;
 	if (*name == ' ')
 		++name;
 
-	return td_outf(level, "%s", name);
+	return strf("%s", name);
+}
+
+int
+pre_do_block(command, p_arg)
+int command;
+char **p_arg;
+{
+	return check_single_expr(command == MR_BLOCK_ADD_EXPRESSION);
 }
 
 /*  Process the return from a block header menu.
  */
 void
-do_block(par, command)
+do_block(par, command, arg)
 objid_t par;
 int command;
+char *arg;
 {
 	fil_t *fil;
 	stackblock_t *sb;
 
 	sb = (stackblock_t *)par;
 	switch(command) {
-	case MR_ADD_VARS:
+	case MR_BLOCK_EXPAND:
 		expand_block(par, sb->sb_block);
 		break;
-	case MR_HIDE_VARS:
+	case MR_BLOCK_COLLAPSE:
 		remove_object(par, OBJ_DESCENDENTS);
 		break;
-	case MR_DISPLAY_SOURCE:
-		fil = sb->sb_stk->stk_func->fu_fil;
+	case MR_BLOCK_DELETE:
+		remove_object(par, OBJ_DESCENDENTS);
+		remove_object(par, OBJ_SELF);
+		break;
+	case MR_BLOCK_DISPLAY_SOURCE:
+		fil = sb->sb_stk->stk_fil;
 		if (fil == NULL)
 			errf("No source file known");
 		else
-			show_source(fil, sb->sb_block->bl_start_lnum, TRUE);
+			show_source(fil, sb->sb_block->bl_start_lnum);
 		break;
-	case MR_ADD_EXPRESSION:
-		get_fi_vars(sb->sb_stk->stk_func->fu_fil); /* get globals loaded */
-		add_expr_object(par, sb->sb_block,
-						sb->sb_stk->stk_func->fu_language);
+	case MR_BLOCK_ADD_EXPRESSION:
+		st_get_fi_vars(sb->sb_stk->stk_fil); /* get globals loaded */
+		add_expr_object(par, sb->sb_block, "",
+				sb->sb_stk->stk_func->fu_language,
+				OBJ_LAST_CHILD);
 		break;
 	default:
 		panic("bad cmd in db");
@@ -965,14 +1074,14 @@ objid_t par;
 
 	f = get_stack_func(par, &junk, &junk);
 #ifdef ARCH_MIPS
-	if (f->fu_language != LANG_C) {
+	if (!(f->fu_language == LANG_C || f->fu_language == LANG_CC)) {
 		errf("Sorry, can only show variables of C functions");
 		return;
 	}
 #endif
 	if (f->fu_flags & FU_NOSYM) {
 		errf("Function %s has no variable information",
-								f->fu_name);
+		     f->fu_demangled_name);
 		return;
 	}
 
@@ -980,31 +1089,51 @@ objid_t par;
 	if (blocks == NULL ||
 	    (blocks->bl_vars == NULL && blocks->bl_blocks == NULL)) {
 		errf("Function %s has no parameters or local variables",
-								f->fu_name);
+		     f->fu_demangled_name);
 		return;
 	}
 
 	expand_block(par, blocks);
 }
 
+static bool
+is_non_empty_block(par)
+block_t *par;
+{
+	block_t *bl;
+
+	if (par->bl_vars != NULL)
+		return TRUE;
+
+	for (bl = par->bl_blocks; bl != NULL; bl = bl->bl_next) {
+		if (is_non_empty_block(bl))
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
 static void
 expand_block(par, parblock)
 objid_t par;
 block_t *parblock;
 {
-	block_t *block;
+	block_t *bl;
 	var_t *v;
 
-	for (block = parblock->bl_blocks; block != NULL; block = block->bl_next) {
-		if (block->bl_vars != NULL || block->bl_blocks != NULL) {
-			if (find_block(par, block) == NULL)
-				add_block_object(par, block);
+	for (bl = parblock->bl_blocks; bl != NULL; bl = bl->bl_next) {
+		if (is_non_empty_block(bl)) {
+			objid_t obj;
+			
+			if ((obj = find_block(par, bl)) == NULL)
+				obj = add_block_object(par, bl);
+			expand_block(obj, bl);
 		}
 	}
 
 	for (v = parblock->bl_vars; v != NULL; v = v->va_next)
 		if (find_var(par, v) == NULL)
-			(void) add_var_object(par, v, OBJ_FIRST_CHILD);
+			(void) add_var_object(par, v, OBJ_FIRST_CHILD, FALSE);
 }
 
 func_t *
@@ -1012,9 +1141,9 @@ get_stack_func(obj, p_fp, p_ap)
 objid_t obj;
 taddr_t *p_fp, *p_ap;
 {
-	stack_t *stk;
+	Stack *stk;
 
-	stk = obj_to_stk(obj);
+	stk = obj_to_stack(obj);
 	*p_fp = stk->stk_fp;
 	*p_ap = stk->stk_ap;
 	return stk->stk_func;
@@ -1022,12 +1151,12 @@ taddr_t *p_fp, *p_ap;
 
 /*  Map a stackpos_t argument to a stack frame pointer.
  */
-static stack_t *
+static Stack *
 stackpos_to_stk(pos)
 stackpos_t pos;
 {
 	struct selst *sel;
-	stack_t *stk;
+	Stack *stk;
 
 	switch(pos) {
 	case STK_OUTER:
@@ -1042,7 +1171,7 @@ stackpos_t pos;
 	case STK_SELECTED:
 		if ((sel = get_selection()) != NULL && sel->se_next == NULL &&
 		    get_object_type(sel->se_code) == OT_FUNC)
-			stk =  (stack_t *)sel->se_code;
+			stk =  (Stack *)sel->se_code;
 		else
 			stk =  Current_stack;
 		break;
@@ -1069,19 +1198,70 @@ bool (*func)PROTO((objid_t stack_obj, func_t *f, char *a_args));
 char *args;
 stackpos_t startpos, limpos;
 {
-	stack_t *stk, *lim;
+	Stack *stk, *lim;
+	bool found = FALSE;
 
 	stk = stackpos_to_stk(startpos);
 	lim = stackpos_to_stk(limpos);
 
 	while (stk != NULL) {
+	  	if (stop_pressed(0, 0))
+		{
+		  found = TRUE;
+		  break;
+		}
 		if ((*func)((objid_t)stk, stk->stk_func, args))
-			return TRUE;
+		{
+		  found = TRUE;
+		  break;
+		}
 		if (stk == lim)
 			break;
 		stk = stk->stk_outer;
 	}
-	return FALSE;
+	return found;
+}
+
+bool
+back_iterate_over_stack_files(sw_id, name, startpos, limpos)
+     int sw_id;
+     char *name;
+     stackpos_t startpos, limpos;
+{
+  Stack *stk, *stk1, *stk_start, *lim;
+  int skip;
+  
+  stk_start = stk = stackpos_to_stk(startpos);
+  lim = stackpos_to_stk(limpos);
+  
+  while (stk != NULL)
+  {
+    if (stk->stk_func->fu_fil)
+    {
+      skip = 0;
+      /* check if we have already done this file */
+      for (stk1 = stk_start; stk1 != NULL;)
+      {
+	if (stk1 == stk)
+	  break;
+	if (stk->stk_func->fu_fil == stk1->stk_func->fu_fil)
+	{
+	  skip = 1;
+	  break;	
+	}
+	stk1 = stk1->stk_inner;
+     }
+      if (!skip && show_global
+	  (sw_id, stk->stk_func->fu_fil, (fil_t*)NULL,
+	   (common_block_t *)NULL,
+	   name, TRUE, FALSE, FALSE, (objid_t *)NULL, TRUE))
+	return TRUE;
+    }
+    if (stk == lim)
+      break;
+    stk = stk->stk_inner;
+  }
+  return FALSE;
 }
 
 /*  Find the address at which the register variable has been saved.
@@ -1089,33 +1269,143 @@ stackpos_t startpos, limpos;
  *  or block).
  */
 taddr_t
-get_reg_addr(par, reg)
+get_reg_addr(xp, par, reg, size)
+target_t *xp;
 objid_t par;
 int reg;
+size_t size;
 {
 	taddr_t addr;
-	stack_t *stk;
+	Stack *stk;
 
-	stk = obj_to_stk(par);
+	stk = obj_to_stack(par);
 
 	switch (reg) {
-	case REG_PC:
+	case UPSREG_PC:
 		addr = (taddr_t)&stk->stk_pc;
 		break;
-	case REG_AP:
+	case UPSREG_AP:
 		addr = (taddr_t)&stk->stk_ap;
 		break;
-	case REG_FP:
+	case UPSREG_FP:
 		addr = (taddr_t)&stk->stk_fp;
 		break;
-	case REG_SP:
+	case UPSREG_SP:
 		addr = (taddr_t)&stk->stk_sp;
 		break;
 	default:
-		if ((addr = reg_addr(stk->stk_inner, reg)) == 0)
+		if ((addr = xp_get_reg_addr(xp, stk, reg)) == 0)
 			addr = regno_to_addr(reg);
+		else
+			addr = adjust_saved_reg_addr(xp, addr, size);
 		break;
 	}
 
 	return addr;
 }
+int match_name_end_in_func(name, obj_exists, obj_var)
+char *name;
+int *obj_exists;
+int **obj_var;
+{
+  objid_t par, var_obj;
+  block_t *block;
+  var_t *v;
+  int found = 0, len, str_match;
+  func_t *f;
+  block_t *blocks;
+  taddr_t junk;
+
+  *obj_exists = 0;
+  *obj_var = 0;
+  if (!can_get_target_vars(get_current_target()))
+    return found;
+
+  par = (objid_t) Current_stack;
+  f = get_stack_func(par, &junk, &junk);
+  
+#ifdef ARCH_MIPS
+  if (!(f->fu_language == LANG_C || f->fu_language == LANG_CC)) {
+    return found;
+  }
+#endif
+  if (f->fu_flags & FU_NOSYM) {
+    return found;
+  }
+  
+  blocks = FU_BLOCKS(f);
+  if (blocks == NULL ||
+      (blocks->bl_vars == NULL && blocks->bl_blocks == NULL)) {
+    return found;
+  }
+  
+  for (block = blocks->bl_blocks; block != NULL; block = block->bl_next) {
+    if (block->bl_vars != NULL || block->bl_blocks != NULL) {
+      if (find_block(par, block) == NULL)
+	add_block_object(par, block);
+    }
+  }
+  len = strlen(name);  
+  for (v = blocks->bl_vars; v != NULL; v = v->va_next)
+  {
+    if (demangling_enabled(0, 0))
+      str_match = !strcmp(v->va_name, name);     
+    else
+      str_match = !!strstr(v->va_name, name);
+    if (str_match)
+    {
+      var_obj = find_var(par, v);
+      if (var_obj)
+	*obj_exists = 1;
+      else
+	var_obj = add_var_object(par, v, OBJ_FIRST_CHILD, FALSE);
+      *obj_var = (int *)var_obj;
+      clear_selection();
+      select_object(var_obj, TRUE, OBJ_SELF);
+      ensure_visible(var_obj);
+      found = 1;
+      break;
+    }
+   }
+  return found;
+}
+
+int
+display_var_in_mangled_form(buf, this_obj_exists, this_obj_var)
+     const char *buf;
+     int *this_obj_exists;
+     int **this_obj_var;
+{
+  char **names, **nptr;
+  int found = 0, found2 = 0, oldstate, obj_exists, *obj_var, *obj_var1;
+  
+  names = ssplit(buf, ".-> \t");
+  oldstate = td_set_obj_updating(OBJ_UPDATING_OFF);
+  
+  /* look for pre mangled name base: "something___<*nptr>" */
+  /* then look for list of post mangled fields */
+  /* "...-><*nptr>___something-><*(nptr++)>___something" etc */
+  
+  if (match_name_end_in_func
+      ("this", this_obj_exists, this_obj_var))
+  {
+    for (nptr = names; *nptr != NULL; ++nptr)
+      found2 = match_name_start_in_obj
+	(*nptr, 0, &obj_exists, &obj_var1);
+    if (found && !found2)
+    {				/* deselect "this" */
+      deselect(code_to_obj(*this_obj_var));
+    }
+  }
+
+  /* RGA - keep looking in case same name is not used in class */
+  found = match_name_end_in_func(*names, &obj_exists, &obj_var);
+  if (found)
+    for (nptr = names, ++nptr; *nptr != NULL; ++nptr)
+      found = match_name_start_in_obj
+	(*nptr, 0, &obj_exists, &obj_var);
+
+  td_set_obj_updating(oldstate);
+  free((char *)names);
+  return (found || found2);
+}
