@@ -7,7 +7,7 @@
  *  unaltered copy of this message (including the copyright).
  */
 
-char ups_va_expr_c_sccsid[] = "@(#)va_expr.c	1.15 15/9/92 (UKC)";
+char ups_va_expr_c_sccsid[] = "@(#)va_expr.c	1.26 24 May 1995 (UKC)";
 
 #include <mtrprog/ifdefs.h>
 
@@ -20,26 +20,31 @@ char ups_va_expr_c_sccsid[] = "@(#)va_expr.c	1.15 15/9/92 (UKC)";
 #include <mtrprog/alloc.h>
 #include <local/wn.h>
 #include <local/obj/obj.h>
-#include <local/obj/fed.h>
+#include <local/edit/edit.h>
 
 #include "objtypes.h"
 #include "ups.h"
 #include "symtab.h"
+#include "reg.h"
 #include "ci.h"
 #include "ui.h"
 #include "va.h"
+#include "target.h"
+#include "st.h"
 #include "data.h"
 #include "exec.h"	/* for SET_EXPR_RESULT_ADDR */
 #include "obj_stack.h"
+#include "obj_util.h"
 #include "menudata.h"
 #include "va_priv.h"
 #include "tdr.h"
+#include "state.h"
 
 /*  Structure representing a displayed expression
  */
 typedef struct dexprst {
 	parse_id_t de_parse_id;
-	code_id_t de_code_id;
+	machine_t *de_machine;
 	block_t *de_block;
 	type_t *de_type;
 	int de_typelevel;
@@ -51,22 +56,29 @@ typedef struct dexprst {
 	struct dexprst *de_nextfree;	/* for free list management only */
 } dexpr_t;
 
-static void expr_keyfunc PROTO((struct edescst *edesc));
-static int expr_quitfunc PROTO((struct edescst *edesc, int n_tries));
 static void expr_edit PROTO((struct drawst dets));
-static ci_exec_result_t set_expr_result PROTO((code_id_t code_id, taddr_t addr,
+static ci_exec_result_t set_expr_result PROTO((machine_t *ma, taddr_t addr,
 					       taddr_t *args,
 					       int nwords, taddr_t *p_res));
 static const char *exprval_to_string PROTO((dexpr_t *de, ci_exec_result_t res));
 static void expr_draw PROTO((draw_t *dets));
 static dexpr_t *dup_dexpr PROTO((dexpr_t *de));
 static dexpr_t *make_dexpr PROTO((block_t *bl));
-static int compile_dexpr_code PROTO((dexpr_t *de, const char *estr, int *p_cnum));
-static int nowrite PROTO((taddr_t unused_addr, const char *unused_buf,
-								int unused_nbytes));
+static bool compile_dexpr_code PROTO((dexpr_t *de, const char *estr,
+				      int *p_cnum));
+static int read_data PROTO((taddr_t addr, void *buf, size_t nbytes));
+static int nowrite PROTO((taddr_t unused_addr, const void *unused_buf,
+							size_t unused_nbytes));
+static bool expr_quitfunc PROTO((objid_t obj, char *new_text,
+				 const char **p_display_string));
 static bool is_hex_digit PROTO((int c));
+static void inc_exprnum PROTO((Edit_display *display));
+static void dec_exprnum PROTO((Edit_display *display));
+static void adjust_expr_number PROTO((Edit_display *display, bool up));
 
-char Expr_format[] = "%[-]60cx%[-]30cs\n";
+/*char Expr_format[] = "%[-]60cx%[-]30cs\n";*/
+/* RGA use elastic format ... */
+char Expr_format[] = "%[-]60cx%[-]:400cs\n";
 
 /*  These are used to pass back the result of an expression from
  *  the builtin function $set_expr_result() to update_var().
@@ -77,16 +89,14 @@ static value_t Expr_value;
 #define FN_EXPR_STR	0
 #define FN_EXPR_VAL	1
 
-#define MAX_EXPR_LEN	250
-
 fdef_t Expr_fdefs[] = {
-	'x', expr_draw, expr_edit, var_or_expr_getwidth,
-	'\0', NULL, NULL, NULL,
+	{ 'x', expr_draw, expr_edit, var_or_expr_getwidth 	},
+	{ '\0', NULL, NULL, NULL				}
 };
 
 fnamemap_t Expr_fnamemap[] = {
-	"expression",	FN_EXPR_STR,
-	NULL,		0,
+	{ FN_EXPR_STR,	"expression",	FALSE,	expr_quitfunc	},
+	{ 0,		NULL,		FALSE,	NULL		}
 };
 
 ALLOC_NEW_FREE(static,dexpr_t,dexpr,de_nextfree)
@@ -122,48 +132,58 @@ objid_t obj;
  *  Also fully dereference the parent.
  */
 int
-get_member_of_aggr(par, name, p_v, p_addr)
+get_member_of_aggr(par, name, p_v, p_addr, btype)
 objid_t par;
 const char *name;
 var_t **p_v;
 taddr_t *p_addr;
+type_t **btype;
 {
 	var_t *v;
-	type_t *btype;
 	int objtype;
 
 	switch (objtype = get_object_type(par)) {
 	case OT_VAR:
-		btype = get_basetype(((dvar_t *)par)->dv_var->va_type);
+		*btype = get_basetype(((dvar_t *)par)->dv_var->va_type);
 		break;
 	case OT_EXPR:
-		btype = ((dexpr_t *)par)->de_type;
-		if (btype->ty_code == DT_PTR_TO)
-			btype = btype->ty_base;
+		*btype = ((dexpr_t *)par)->de_type;
+
+		if (*btype == NULL) {
+			/*  This happens if we failed to evaluate the
+			 *  parent expression (e.g. when a struct pointer
+			 *  points to an invalid address after a
+			 *  restore_expr().
+			 */
+			return -1;
+		}
+		
+		if ((*btype)->ty_code == DT_PTR_TO)
+			*btype = (*btype)->ty_base;
 		break;
 	default:
 		panic("bad objtype in gma");
-		btype = NULL;	/* to satisfy gcc */
+		*btype = NULL;	/* to satisfy gcc */
 	}
 
-	if (btype->ty_code == TY_STRUCT || btype->ty_code == TY_UNION) {
-		for (v = btype->ty_aggr_or_enum->ae_aggr_members; v != NULL; v = v->va_next) {
+	if ((*btype)->ty_code == TY_STRUCT || (*btype)->ty_code == TY_UNION) {
+		for (v = (*btype)->ty_aggr_or_enum->ae_aggr_members; v != NULL; v = v->va_next) {
 			if (strcmp(v->va_name, name) == 0) {
 				*p_v = v;
 				if (objtype == OT_VAR)
 					deref_aggr((dvar_t *)par);
 				*p_addr = var_or_expr_addr(par);
-				return TRUE;
+				return 0;
 			}
 		}
 	}
-	return FALSE;
+	return -1;
 }
 
 void
 expr_getcolor(obj, p_fg, p_bg)
 objid_t obj;
-short *p_fg, *p_bg;
+long *p_fg, *p_bg;
 {
 	get_value_colors(((dexpr_t *)obj)->de_val_changed, p_fg, p_bg);
 }
@@ -175,8 +195,12 @@ draw_t *dets;
 	const char *s;
 
 	s = (const char *)dets->dr_fval;
-	wn_ttext(dets->dr_wn, s, dets->dr_x, dets->dr_y, dets->dr_fg, dets->dr_bg);
+	  /* RGA use wn_wd_ttext() do expr values rescale */
 	dets->dr_width = get_decl_len(strlen(s)) * wn_get_sysfont()->ft_width;
+	if (dets->selected)
+	  wn_wd_ttext(dets->dr_wn, s, dets->dr_x, dets->dr_y, dets->dr_width, dets->dr_bg, dets->dr_fg);
+	else
+	  wn_wd_ttext(dets->dr_wn, s, dets->dr_x, dets->dr_y, dets->dr_width, dets->dr_fg, dets->dr_bg);
 }
 
 static bool
@@ -187,31 +211,41 @@ int c;
 }
 
 static void
-expr_keyfunc(edesc)
-edesc_t *edesc;
+inc_exprnum(display)
+Edit_display *display;
 {
+	adjust_expr_number(display, TRUE);
+}
+
+static void
+dec_exprnum(display)
+Edit_display *display;
+{
+	adjust_expr_number(display, FALSE);
+}
+
+static void
+adjust_expr_number(display, up)
+Edit_display *display;
+bool up;
+{
+	Edit_buffer *buffer;
+	Field_edit_info *fe;
+	objid_t obj;
+	char *text;
 	const char *pos, *endnum, *fmt;
-	bool up, seen_hex_digit, non_octal;
+	char *new_text;
+	bool seen_hex_digit, non_octal;
 	int radix, inc, dig_offset, i, junk_cnum;
 	long origval;
-	char buf[MAX_EXPR_LEN];
-
-	if (edesc->ed_meaning != EDM_INSERT_CHAR &&
-					edesc->ed_meaning != EDM_ILLEGAL_CHAR)
-		return;
+	size_t point;
 	
-	switch (edesc->ed_char) {
-	case CONTROL('K'):
-		up = FALSE;
-		break;
-	case CONTROL('J'):
-		up = TRUE;
-		break;
-	default:
-		return;
-	}
+	buffer = edit_get_buffer(display);
 
-	pos = edesc->ed_copy + edesc->ed_curpos;
+	text = get_buffer_contents(buffer);
+	point = edit_get_point(display);
+	
+	pos = text + point;
 
 	seen_hex_digit = non_octal = FALSE;
 	for (dig_offset = 0; is_hex_digit(pos[dig_offset]); ++dig_offset) {
@@ -222,7 +256,7 @@ edesc_t *edesc;
 	}
 	endnum = pos + dig_offset;
 
-	while (pos > edesc->ed_copy && is_hex_digit(pos[-1])) {
+	while (pos > text && is_hex_digit(pos[-1])) {
 		--pos;
 		if (*pos == '8' || *pos == '9')
 			non_octal = TRUE;
@@ -232,28 +266,27 @@ edesc_t *edesc;
 	
 	/*  Determine the radix of the number.
 	 */
-	if (pos > edesc->ed_copy + 1 && strncmp(pos - 2, "0x", 2) == 0) {
+	if (pos > text + 1 && strncmp(pos - 2, "0x", 2) == 0) {
 		pos -= 2;
 		radix = 16;
-		fmt = "0x%lx";
+		fmt = "%.*s0x%lx%s";
 	}
 	else if (*pos == '0' && isdigit(pos[1])) {
 		radix = 8;
-		fmt = "0%lo";
+		fmt = "%.*s0%lo%s";
 	}
 	else {
 		radix = 10;
-		fmt = "%ld";
+		fmt = "%.*s%ld%s";
 	}
 
 	if (isalpha(*endnum) || !is_hex_digit(*pos) ||
-		     (seen_hex_digit && radix != 16) || (non_octal && radix == 8)) {
+	    (seen_hex_digit && radix != 16) || (non_octal && radix == 8)) {
 		errf("Not pointing at a number");
-		edesc->ed_meaning = EDM_CONT;
 		return;
 	}
 
-	if (pos > edesc->ed_copy && pos[-1] == '-')
+	if (pos > text && pos[-1] == '-')
 		--pos;
 	
 	inc = 1;
@@ -262,13 +295,12 @@ edesc_t *edesc;
 
 	origval = strtol(pos, (char **)NULL, 0);
 
-	sprintf(buf, "%.*s", pos - edesc->ed_copy, edesc->ed_copy);
-	sprintf(buf + strlen(buf), fmt, origval + (up ? inc : -inc));
-	strcat(buf, endnum);
+	new_text = strf(fmt,
+			pos - text, text, origval + (up ? inc : -inc), endnum);
 	
 	/*  Point pos at the end of the number in the new string.
 	 */
-	pos = buf + (pos - edesc->ed_copy);
+	pos = new_text + (pos - text);
 	if (*pos == '-')
 		++pos;
 	if (radix == 16)
@@ -279,62 +311,62 @@ edesc_t *edesc;
 	/*  Try to maintain the same number of digits as before after
 	 *  the marker bar.
 	 */
-	while (pos > buf && is_hex_digit(pos[-1]) && dig_offset > 0) {
+	while (pos > new_text && is_hex_digit(pos[-1]) && dig_offset > 0) {
 		--dig_offset;
 		--pos;
 	}
 
-	edesc->ed_curpos = pos - buf;
-	ed_newstring(edesc, buf);
+	point = pos - new_text;
+	
+	wn_updating_off(WN_STDWIN);
+	
+	edit_delete(buffer, 0, edit_get_buffer_length(buffer));
+	edit_insert(buffer, 0, new_text, strlen(new_text), TRUE);
 
-	wn_updating_off(edesc->ed_wn);
-	ed_cursor(edesc, EDC_CURSOR_OFF);
-	change_field((objid_t)edesc->ed_user, FN_EXPR_STR, edesc->ed_copy);
-	ed_new_orig(edesc, get_field_value((objid_t)edesc->ed_user, FN_EXPR_STR));
+	fe = (Field_edit_info *)edit_get_user_data(display);
+	obj = field_edit_get_object(fe);
+	
+	change_field(obj, FN_EXPR_STR, new_text);
+	field_edit_update_orig_text(fe, new_text);
+	
+	compile_dexpr_code((dexpr_t *)obj, new_text, &junk_cnum);
+	update_expr(obj, DF_NONE, TRUE);
 
-	compile_dexpr_code((dexpr_t *)edesc->ed_user, edesc->ed_copy, &junk_cnum);
-	update_expr((objid_t)edesc->ed_user, TRUE);
+	edit_move_point(display, point);
 
-	ed_cursor(edesc, EDC_CURSOR_ON);
-	wn_updating_on(edesc->ed_wn);
+	wn_updating_on(WN_STDWIN);
 
-	edesc->ed_meaning = EDM_CONT;
+	free(new_text);
 }
 
-static int
-expr_quitfunc(edesc, n_tries)
-struct edescst *edesc;
-int n_tries;
+static bool
+expr_quitfunc(obj, new_text, p_display_string)
+objid_t obj;
+char *new_text;
+const char **p_display_string;
 {
 	dexpr_t *de;
-	int cnum, res;
+	int cnum;
 
-	if (strcmp(edesc->ed_orig, edesc->ed_copy) == 0)
-		return EDR_CONFIRM_NO_CHANGE;
-
-	if (message_on_null_name(edesc, "expression") == -1)
-		return force_quit(n_tries);
-
-	de = (dexpr_t *)edesc->ed_user;
-	if (compile_dexpr_code(de, edesc->ed_copy, &cnum) != 0) {
-		res = force_quit(n_tries);
-		if (cnum != -1 && res != EDR_CANCEL_AND_QUIT) {
-			edesc->ed_meaning = EDM_SETCURSOR;
-			edesc->ed_newpos = cnum;
-			do_edit(edesc, EDM_SETCURSOR);
-		}
-		return res;
+	de = (dexpr_t *)obj;
+	
+	if (!compile_dexpr_code(de, new_text, &cnum)) {
+		/* TODO: do something with cnum */
+		change_field((objid_t)de, FN_EXPR_VAL, "<error>");
+		return FALSE;
 	}
 
 	td_set_obj_updating(OBJ_UPDATING_OFF);
-	change_field((objid_t)edesc->ed_user, FN_EXPR_STR, edesc->ed_copy);
-	update_expr((objid_t)de, TRUE);
+	change_field((objid_t)de, FN_EXPR_STR, new_text);
+	update_expr((objid_t)de, DF_NONE, TRUE);
 	td_set_obj_updating(OBJ_UPDATING_ON);
 
-	return EDR_CONFIRM_CHANGE;
+	note_var_change((objid_t)de);
+
+	return TRUE;
 }
 
-static int
+static bool
 compile_dexpr_code(de, estr, p_cnum)
 dexpr_t *de;
 const char *estr;
@@ -349,51 +381,48 @@ int *p_cnum;
 
 	if (cr->cr_code_has_func_calls) {
 		errf("Can't call target functions in expressions");
-		free_parse_and_code_ids(cr->cr_parse_id, cr->cr_code_id);
+		free_parse_id_and_machine(cr->cr_parse_id, cr->cr_machine);
 		*p_cnum = -1;
-		return -1;
+		return FALSE;
 	}
 
-	if (cr->cr_code_id == NULL || cr->cr_parse_id == NULL) {
-		free_parse_and_code_ids(cr->cr_parse_id, cr->cr_code_id);
+	if (cr->cr_machine == NULL || cr->cr_parse_id == NULL) {
+		free_parse_id_and_machine(cr->cr_parse_id, cr->cr_machine);
 		*p_cnum = lxbuf.lx_cnum;
-		return -1;
+		return FALSE;
 	}
 
-	free_parse_and_code_ids(de->de_parse_id, de->de_code_id);
+	free_parse_id_and_machine(de->de_parse_id, de->de_machine);
 
-	de->de_code_id = cr->cr_code_id;
+	de->de_machine = cr->cr_machine;
 	de->de_parse_id = cr->cr_parse_id;
 
-	return 0;
+	return TRUE;
 }
 
 static void
 expr_edit(fdets)
 struct drawst fdets;
 {
-	dexpr_t *de;
-	edesc_t edescbuf;
+	Edit_display *display;
+	Edit_keymap *keymap;
 
-	de = (dexpr_t *)fdets.dr_code;
+	display = field_edit_start(&fdets, "expression", (char *)NULL);
 
-	clear_selection();
-	make_edesc(&edescbuf, fdets.dr_wn, (const char *)fdets.dr_fval,
-						MAX_EXPR_LEN,
-						fdets.dr_fg, fdets.dr_bg);
-
-	/*  BUG: must think of a better way to do this.
+	/*  If we are replaying events there is no display object.
 	 */
-	if (fdets.dr_fval[0] == '\0' && fdets.dr_wn != -1) {
-		edescbuf.ed_meaning = EDM_SETCURSOR;
-		edescbuf.ed_newpos = 0;
-		do_edit(&edescbuf, EDM_SETCURSOR);
-	}
+	if (display == NULL)
+		return;
 
-	edescbuf.ed_user = (int)de;
-	edescbuf.ed_quitfunc = (ed_quitfunc_t)expr_quitfunc;
-	edescbuf.ed_keyfunc = (ed_keyfunc_t)expr_keyfunc;
-	suppress_ta_cursor_then_edit_field(&edescbuf, "expression");
+	keymap = edit_create_keymap(edit_get_keymap(display));
+
+	edit_add_keymap_control_entry(keymap, 'n', inc_exprnum);
+	edit_add_keymap_entry(keymap, 0, EDIT_DOWN_ARROW, inc_exprnum);
+	
+	edit_add_keymap_control_entry(keymap, 'p', dec_exprnum);
+	edit_add_keymap_entry(keymap, 0, EDIT_UP_ARROW, dec_exprnum);
+
+	edit_set_keymap(display, keymap);
 }
 
 const char *
@@ -407,6 +436,9 @@ objid_t obj;
 
 	name = get_field_value(obj, FN_EXPR_STR);
 
+	if (*name == '\0')
+		name = "[BLANK]";
+
 	count = 1;
 	obj2 = get_code(get_code(obj, OBJ_PARENT), OBJ_CHILD);
 
@@ -431,27 +463,42 @@ objid_t obj;
 	return last;
 }
 
-int
-expr_dumpobj(arg, code, level)
-char *arg;
+char *
+expr_format_obj(code)
 objid_t code;
-int level;
 {
-	return td_outf(level, "%20s %s", get_field_value(code, FN_EXPR_STR),
+	return strf("%20s %s", get_field_value(code, FN_EXPR_STR),
 					 get_field_value(code, FN_EXPR_VAL));
 }
 
+int
+check_single_expr(is_add_expr_command)
+bool is_add_expr_command;
+{
+	if (is_add_expr_command && get_num_selected() != 1) {
+		errf("Can only add expressions one at a time");
+		return -1;
+	}
+
+	return 0;
+}
+
 /*  Process the return from a var menu.
  */
 void
-do_expr(obj, cmd)
+do_expr(obj, cmd, arg)
 objid_t obj;
 int cmd;
+char *arg;
 {
 	dexpr_t *de;
 
 	de = (dexpr_t *)obj;
 
+	if ( cmd == MR_VAR_ACCELERATOR_ACTION )
+	  cmd = va_accelerator_action_from_type(de->de_type, de->de_format);
+
+
 	switch(cmd) {
 	case MR_VAR_STRING:
 	    {
@@ -475,11 +522,10 @@ int cmd;
 
 		de->de_format = DF_STRING;
 		oldstate = td_set_obj_updating(OBJ_UPDATING_OFF);
-		update_expr(obj, TRUE);
+		update_expr(obj, DF_NONE, TRUE);
 		td_set_obj_updating(oldstate);
 		break;
 	    }
-		break;
 	case MR_VAR_SIGNED_DECIMAL:
 	case MR_VAR_UNSIGNED_DECIMAL:
 	case MR_VAR_SIGNED_OCTAL:
@@ -488,12 +534,19 @@ int cmd;
 	case MR_VAR_UNSIGNED_HEX:
 	case MR_VAR_UNSIGNED_BINARY:
 	case MR_VAR_ASCII_BYTE:
+        case MR_VAR_DEFAULT_FORMAT:
 	    {
 		int oldstate;
 
 		de->de_format = mval_to_vformat(cmd);
+
+		if (de->de_format == DF_NONE) {
+			de->de_format = default_format(de->de_type,
+						       de->de_type);
+		}
+		
 		oldstate = td_set_obj_updating(OBJ_UPDATING_OFF);
-		update_expr(obj, TRUE);
+		update_expr(obj, DF_NONE, TRUE);
 		td_set_obj_updating(oldstate);
 		break;
 	    }
@@ -503,7 +556,10 @@ int cmd;
 	case MR_VAR_ADDRESS:
 		errf("Can't decrease the indirection level of an expression");
 		break;
-	case MR_VAR_EXPAND:
+	case MR_VAR_EXPAND_LIKE_BEFORE:
+		errf("elb NYI");
+		break;
+	case MR_VAR_EXPAND_COMPLETELY:
 		if (de->de_type == NULL) {
 			errf("Can't expand illegal expression");
 			break;
@@ -521,7 +577,8 @@ int cmd;
 			v = de->de_type->ty_base->ty_aggr_or_enum->ae_aggr_members;
 			for (; v != NULL; v = v->va_next) {
 				if (find_var((objid_t)de, v) == NULL)
-					add_var_object(obj, v, OBJ_FIRST_CHILD);
+					add_var_object(obj, v, OBJ_FIRST_CHILD,
+						       FALSE);
 			}
 			break;
 		    }
@@ -539,17 +596,22 @@ int cmd;
 		break;
 	case MR_VAR_DUP:
 		add_to_new_selection((objid_t)dup_dexpr(de));
+	    	note_var_change((objid_t)de);
 		break;
 	case MR_VAR_DELETE:
+		note_var_change(obj);
 		remove_object(obj, OBJ_DESCENDENTS);
 		remove_object(obj, OBJ_SELF);
 		break;
 	case MR_VAR_COLLAPSE:
+		note_var_change(obj);
 		remove_object(obj, OBJ_CHILDREN);
 		break;
 	case MR_VAR_COLLAPSE_COMPLETELY:
+		note_var_change(obj);
 		remove_object(obj, OBJ_DESCENDENTS);
 		break;
+	case MR_VAR_RWCSTRING:
 	case MR_VAR_WANT_TYPEDEFS:
 	case MR_VAR_NO_TYPEDEFS:
 		/*  Typedefs don't apply to expressions, as the declaration
@@ -561,6 +623,60 @@ int cmd;
 	}
 }
 
+void
+get_expr_info(obj, p_bl, p_text, p_format)
+objid_t obj;
+block_t **p_bl;
+const char **p_text;
+vformat_t *p_format;
+{
+	dexpr_t *de;
+
+	de = (dexpr_t *)obj;
+
+	*p_bl = de->de_block;
+	*p_text = get_field_value(obj, FN_EXPR_STR);
+	*p_format = de->de_format;
+}
+
+type_t *
+get_expr_type(obj)
+objid_t obj;
+{
+	return ((dexpr_t *)obj)->de_type;
+}
+
+objid_t
+restore_displayed_expr(par, bl, text, format)
+objid_t par;
+block_t *bl;
+const char *text;
+vformat_t format;
+{
+	dexpr_t *newde;
+	int junk_cnum;
+
+	newde = make_dexpr(bl);
+
+	/*  Note that we must create the new object before calling
+	 *  compile_dexpr_code because that eventually calls
+	 *  get_regaddr_for_ci() which needs the object to exist.
+	 */
+	new_object((objid_t)newde, OT_EXPR, par, OBJ_FIRST_CHILD);
+
+	if (!compile_dexpr_code(newde, text, &junk_cnum)) {
+		remove_object((objid_t)newde, OBJ_SELF);
+		return NULL;
+	}
+
+	set_field_value((objid_t)newde, FN_EXPR_STR, (fval_t)strsave(text));
+	set_field_value((objid_t)newde, FN_EXPR_VAL, (fval_t)strsave("0"));
+
+	update_expr((objid_t)newde, format, TRUE);
+
+	return (objid_t)newde;
+}
+
 static dexpr_t *
 dup_dexpr(de)
 dexpr_t *de;
@@ -578,21 +694,13 @@ dexpr_t *de;
 	new_object((objid_t)newde, OT_EXPR, (objid_t)de, OBJ_AFTER);
 
 	estr = (char *)get_field_value((objid_t)de, FN_EXPR_STR);
-	if (compile_dexpr_code(newde, estr, &junk_cnum) != 0)
+	if (!compile_dexpr_code(newde, estr, &junk_cnum))
 		panic("cdc botch in dd");
 
 	set_field_value((objid_t)newde, FN_EXPR_STR, (fval_t)strsave(estr));
 	set_field_value((objid_t)newde, FN_EXPR_VAL, (fval_t)strsave("0"));
 
-	/*  We want the same format for the new expression as for the old.
-	 *  update_expr() will reset the format to the default if it sees
-	 *  a change in basecode or typelevel so copy these as well.
-	 */
-	newde->de_basecode = de->de_basecode;
-	newde->de_typelevel = de->de_typelevel;
-	newde->de_format = de->de_format;
-
-	update_expr((objid_t)newde, TRUE);
+	update_expr((objid_t)newde, de->de_format, TRUE);
 
 	return newde;
 }
@@ -606,7 +714,7 @@ block_t *bl;
 	de = new_dexpr();
 
 	de->de_parse_id = NULL;
-	de->de_code_id = NULL;
+	de->de_machine = NULL;
 	de->de_block = bl;
 	de->de_type = NULL;
 	de->de_basecode = TY_NOTYPE;
@@ -619,33 +727,60 @@ block_t *bl;
 	return de;
 }
 
-void
-add_expr_object(par, bl, language)
+objid_t
+add_expr_object(par, bl, text, language, poscode)
 objid_t par;
 block_t *bl;
+const char *text;
 language_t language;
+int poscode;
 {
 	dexpr_t *de;
 	objid_t obj;
+	int junk_cnum;
 
-	if (language != LANG_C) {
-		errf("Sorry, can only add expressions in C");
-		return;
+	/* RGA allow expns in C++ header files of type ".h" - if
+	 type is ".H" or ".hh", language will be LANG_CC */
+
+	if (!(language == LANG_C || language == LANG_CC
+	      || get_user_lang() == LANG_CC
+	      || (language == LANG_UNKNOWN &&
+		  get_mainfunc_lang() == LANG_CC))) {
+		errf("Sorry, can only add expressions in C or C++");
+		return NULL;
 	}
 
 	de = make_dexpr(bl);
 	obj = (objid_t)de;
 
 	td_set_obj_updating(OBJ_UPDATING_OFF);
-	new_object(obj, OT_EXPR, par, OBJ_FIRST_CHILD);
-	set_field_value(obj, FN_EXPR_STR, (fval_t)strsave(""));
+	
+	new_object(obj, OT_EXPR, par, poscode);
+	set_field_value(obj, FN_EXPR_STR, (fval_t)strsave(text));
 	set_field_value(obj, FN_EXPR_VAL, (fval_t)strsave("0"));
+	
+	if (*text != '\0') {
+		de->de_format = DF_NONE;
+		
+		if (!compile_dexpr_code(de, text, &junk_cnum)) {
+			change_field((objid_t)de, FN_EXPR_VAL, "<error>");
+			text = NULL;
+		}
+	}
+	
 	ensure_visible(obj);
+
 	td_set_obj_updating(OBJ_UPDATING_ON);
 
-	td_obj_edit_field(obj, FN_EXPR_STR, 0, 0);
-	if (de->de_code_id == NULL)
-		remove_object(obj, OBJ_SELF);
+	if (text == NULL || *text == '\0') {
+		edit_field_obj(obj, FN_EXPR_STR);
+	}
+	else {
+		update_expr(obj, de->de_format, FALSE);
+		select_object(obj, TRUE, OBJ_SELF);
+	}
+
+	return obj;
 }
 
 void
@@ -655,22 +790,32 @@ objid_t obj;
 	dexpr_t *de;
 
 	de = (dexpr_t *)obj;
-	free_parse_and_code_ids(de->de_parse_id, de->de_code_id);
+	free_parse_id_and_machine(de->de_parse_id, de->de_machine);
 	free_dexpr((dexpr_t *)obj);
 }
 
 static int
 nowrite(unused_addr, unused_buf, unused_nbytes)
 taddr_t unused_addr;
-const char *unused_buf;
-int unused_nbytes;
+constvoidptr unused_buf;
+size_t unused_nbytes;
 {
 	return -1;
 }
 
+static int
+read_data(addr, buf, nbytes)
+taddr_t addr;
+void *buf;
+size_t nbytes;
+{
+	return dread(get_current_target(), addr, buf, nbytes);
+}
+
 void
-update_expr(obj, change_caused_by_edit)
+update_expr(obj, new_format, change_caused_by_edit)
 objid_t obj;
+vformat_t new_format;
 bool change_caused_by_edit;
 {
 	taddr_t fp, ap;
@@ -685,7 +830,7 @@ bool change_caused_by_edit;
 
 	de = (dexpr_t *)obj;
 
-	if (de->de_code_id == NULL) {
+	if (de->de_machine == NULL) {
 		change_field(obj, FN_EXPR_VAL, "<syntax error>");
 
 		/*  This change is always caused by an edit.
@@ -702,10 +847,11 @@ bool change_caused_by_edit;
 		get_stack_func(par, &fp, &ap);
 
 	Expr_type = NULL;
-	ci_initialise_code(de->de_code_id, TRUE);
+	ci_initialise_machine(de->de_machine, TRUE, FALSE,
+					(char **)NULL, (char **)NULL);
 
-	res = ci_execute_code(de->de_code_id, fp, ap,
-					dread, nowrite, set_expr_result);
+	res = ci_execute_machine(de->de_machine, fp, ap,
+					read_data, nowrite, set_expr_result);
 
 	if (res != CI_ER_TRAP) {
 		switch (res) {
@@ -730,6 +876,8 @@ bool change_caused_by_edit;
 	if (basecode != de->de_basecode || de->de_typelevel != typelevel) {
 		if (Expr_type == NULL)
 			de->de_format = DF_NONE;
+		else if (new_format != DF_NONE)
+			de->de_format = new_format;
 		else if (Expr_type->ty_code == DT_PTR_TO &&
 		         Expr_type->ty_base->ty_code == TY_CHAR)
 			de->de_format = DF_STRING;
@@ -769,6 +917,8 @@ bool change_caused_by_edit;
 	
 	if (de->de_aggrtype != NULL)
 		update_struct(obj, change_caused_by_edit);
+
+	note_var_change(obj);
 }
 
 	
@@ -780,9 +930,11 @@ dexpr_t *de;
 ci_exec_result_t res;
 {
 	value_t vl;
+	target_t *xp;
 	type_t *type;
 	static char buf[128];
 
+	xp = get_current_target();
 	type = de->de_type;
 	vl = de->de_value;
 
@@ -791,10 +943,8 @@ ci_exec_result_t res;
 		break;
 	case CI_ER_WRITEDATA_FAILED:
 		return "<illegal data modification>";
-		break;
 	case CI_ER_READDATA_FAILED:
 		return "<bad address>";
-		break;
 	default:
 		strnf(buf, sizeof(buf), "<%s>", ci_exec_result_to_string(res));
 		return buf;
@@ -811,18 +961,28 @@ ci_exec_result_t res;
 		break;
 	case TY_INT:
 	case TY_LONG:
-		int_to_string(buf, sizeof(buf), vl.vl_int, de->de_format);
+	case TY_LONGLONG:
+		int_to_string(buf, sizeof(buf), (vlong)vl.vl_int, de->de_format);
 		break;
 	case TY_UINT:
 	case TY_ULONG:
-		int_to_string(buf, sizeof(buf), vl.vl_int, de->de_format);
+	case TY_ULONGLONG:
+		int_to_string(buf, sizeof(buf), (vlong)vl.vl_int, de->de_format);
 		break;
 	case TY_FLOAT:
-		strcpy(buf, get_real(vl, FALSE, TRUE));
+		strcpy(buf, get_real(xp->xp_words_big_endian, vl, FALSE,
+				     sizeof (float)));
 		break;
 	case TY_DOUBLE:
-		strcpy(buf, get_real(vl, FALSE, FALSE));
+		strcpy(buf, get_real(xp->xp_words_big_endian, vl, FALSE,
+				     sizeof (double)));
 		break;
+#if (defined(__GNUC__) && !defined (OS_SUNOS_4)) || defined(HAVE_LONG_DOUBLE)
+ 	case TY_LONGDOUBLE:
+ 		strcpy(buf, get_real(xp->xp_words_big_endian, vl, FALSE,
+ 				     sizeof (long double)));
+		break;
+#endif
 	case TY_ENUM:
 	case TY_U_ENUM:
 		enumval_to_string(buf, sizeof(buf), vl.vl_int, type);
@@ -835,8 +995,8 @@ ci_exec_result_t res;
 }
 
 static ci_exec_result_t
-set_expr_result(code_id, addr, args, nwords, p_res)
-code_id_t code_id;
+set_expr_result(ma, addr, args, nwords, p_res)
+machine_t *ma;
 taddr_t addr;
 taddr_t *args;
 int nwords;
@@ -860,7 +1020,8 @@ taddr_t *p_res;
 		break;
 	case TY_DOUBLE:
 		argslots = sizeof(double) / sizeof(long);
-		Expr_value.vl_double = *(double *)args;
+		memcpy((char *)&Expr_value.vl_double, (char *)args,
+		       sizeof(double));
 		break;
 	case DT_ARRAY_OF:
 	case DT_PTR_TO:
